# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Registry-like monster"""

__docformat__ = 'restructuredtext'

from ConfigParser import SafeConfigParser
import os.path
from os.path import join as opj


class ConfigManager(SafeConfigParser):
    """Central configuration registry for testkraut.

    The purpose of this class is to collect all configurable settings used by
    various parts of testkraut. It is fairly simple and does only little more
    than the standard Python ConfigParser. Like ConfigParser it is blind to the
    data that it stores, i.e. not type checking is performed.

    Configuration files (INI syntax) in multiple location are passed when the
    class is instantiated or whenever `Config.reload()` is called later on.
    By default it looks for a config file named `testkraut.cfg` in the current
    directory and `.testkraut.cfg` in the user's home directory. Moreover, the
    constructor takes an optional argument with a list of additional file names
    to parse.

    In addition to configuration files, this class also looks for special
    environment variables to read settings from. Names of such variables have to
    start with `TESTKRAUT_` following by the an optional section name and the
    variable name itself ('_' as delimiter). If no section name is provided,
    the variables will be associated with section `general`. Some examples::

        TESTKRAUT_VERBOSE=1

    will become::

        [general]
        verbose = 1

    However, `TESTKRAUT_VERBOSE_OUTPUT=stdout` becomes::

        [verbose]
        output = stdout

    Any length of variable name as allowed, e.g. TESTKRAUT_SEC1_LONG_NAME=1
    becomes::

        [sec1]
        long name = 1

    Settings from custom configuration files (specified by the constructor
    argument) have the highest priority and override settings found in the
    current directory. They in turn override user-specific settings and finally
    the content of any `TESTKRAUT_*` environment variables overrides all
    settings read from any file.
    """

    # things we want to count on to be available
    _DEFAULTS = {'general':
                  {
                    'verbose': '1',
                  }
                }


    def __init__(self, filenames=None):
        """Initialization reads settings from config files and env. variables.

        Parameters
        ----------
        filenames : list of filenames
        """
        SafeConfigParser.__init__(self)

        # store additional config file names
        if not filenames is None:
            self.__cfg_filenames = filenames
        else:
            self.__cfg_filenames = []

        # set critical defaults
        for sec, vars in ConfigManager._DEFAULTS.iteritems():
            self.add_section(sec)
            for key, value in vars.iteritems():
                self.set(sec, key, value)

        # now get the setting
        self.reload()


    def reload(self):
        """Re-read settings from all configured locations.
        """
        # listof filenames to parse (custom plus some standard ones), they are
        # listed in the order they need to be parsed to guarantee a sane
        # configuration file cascade
        homedir = os.path.expanduser('~')
        cfg_file_candidates = [
                # the base config shipped in the package
                opj(os.path.dirname(__file__), 'testkraut.cfg'),
                # systemwide config
                '/etc/testkraut/testkraut.cfg',
                # user config
                opj(homedir, '.testkraut.cfg'),
                # current dir config
                'testkraut.cfg'
            ]

        # read local and user-specific config
        files = self.read(cfg_file_candidates + self.__cfg_filenames)

        # no look for variables in the environment
        for var in [v for v in os.environ.keys() if v.startswith('TESTKRAUT_')]:
            # strip leading 'TESTKRAUT_' and lower case entries
            svar = var[10:].lower()

            # section is next element in name (or 'general' if simple name)
            if not svar.count('_'):
                sec = 'general'
            else:
                cut = svar.find('_')
                sec = svar[:cut]
                svar = svar[cut + 1:].replace('_', ' ')

            # check if section is already known and add it if not
            if not self.has_section(sec):
                self.add_section(sec)

            # set value
            self.set(sec, svar, os.environ[var])


    def __repr__(self):
        """Generate INI file content with current configuration.
        """
        # make adaptor to use str as file-like (needed for ConfigParser.write()
        class file2str(object):
            def __init__(self):
                self.__s = ''
            def write(self, val):
                self.__s += val
            def str(self):
                return self.__s

        r = file2str()
        self.write(r)

        return r.str()


    def save(self, filename):
        """Write current configuration to a file.
        """
        f = open(filename, 'w')
        self.write(f)
        f.close()


    def get(self, section, option, default=None, **kwargs):
        """Wrapper around SafeConfigParser.get() with a custom default value.

        This method simply wraps the base class method, but adds a `default`
        keyword argument. The value of `default` is returned whenever the
        config parser does not have the requested option and/or section.
        """
        if not self.has_option(section, option):
            return default

        try:
            return SafeConfigParser.get(self, section, option, **kwargs)
        except ValueError, e:
            # provide somewhat descriptive error
            raise ValueError, \
                  "Failed to obtain value from configuration for %s.%s. " \
                  "Original exception was: %s" % (section, option, e)


    def getboolean(self, section, option, default=None):
        """Wrapper around SafeConfigParser.getboolean() with a custom default.

        This method simply wraps the base class method, but adds a `default`
        keyword argument. The value of `default` is returned whenever the
        config parser does not have the requested option and/or section.
        """
        if not self.has_option(section, option):
            if isinstance(default, bool):
                return default
            else:
                # compatibility layer for py3 version of ConfigParser
                if hasattr(self, '_boolean_states'):
                    boolean_states = self._boolean_states
                else:
                    boolean_states = self.BOOLEAN_STATES
                if default.lower() not in boolean_states:
                    raise ValueError, 'Not a boolean: %s' % default
                return boolean_states[default.lower()]

        return SafeConfigParser.getboolean(self, section, option)


    ##REF: Name was automagically refactored
    def get_as_dtype(self, section, option, dtype, default=None):
        """Convenience method to query options with a custom default and type

        This method simply wraps the base class method, but adds a `default`
        keyword argument. The value of `default` is returned whenever the
        config parser does not have the requested option and/or section.

        In addition, the returned value is converted into the specified `dtype`.
        """
        if not self.has_option(section, option):
            return default
        try:
            return SafeConfigParser._get(self, section, dtype, option)
        except ValueError, e:
            # provide somewhat descriptive error
            raise ValueError, \
                  "Failed to obtain value from configuration for %s.%s. " \
                  "Original exception was: %s" % (section, option, e)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

import json
import difflib
import numpy as np
from uuid import uuid1 as uuid

__allowed_spec_keys__ = [
        'dependencies',
        'description',
        'entities',
        'environment',
        'evaluations',
        'executables',
        'id',
        'inputs',
        'outputs',
        'processes',
        'system',
        'test',
        'version',
    ]

def _raise(exception, why, input=None):
    if not input is None:
        input = ' (got: %s)' % repr(input)
    else:
        input = ''
    raise exception("SPEC: %s%s" % (why, input))

def _verify_tags(struct, tags, name):
    for tag in tags:
        if not tag in struct:
            _raise(ValueError,
                        "mandatory key '%s' is not in %s" % (tag, name))

def _verify_spec_tags(specs, tags, name):
    for i, os_id in enumerate(specs):
        os = specs[os_id]
        _verify_tags(os, tags, '%s: %s' % (name, os_id))

class SPECJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, np.ndarray):
            return list(o)
        return super(SPECJSONEncoder, self).default(o)

class SPEC(dict):
    def __init__(self, src=None):
        dict.__init__(self)
        if isinstance(src, file):
            self.update(json.load(src))
        elif isinstance(src, basestring):
            self.update(json.loads(src))
        elif isinstance(src, dict):
            self.update(src)
        # charge with sane defaults
        if not 'id' in self:
            self['id'] = uuid().hex
        if not 'version' in self:
            self['version'] = 0
        self._check()

    def _check(self):
        _verify_tags(self, ('id', 'version', 'test'), 'SPEC')
        for i, ev in enumerate(self.get('evaluations', {})):
            _verify_tags(ev, ('id', 'input spec', 'operator'),
                         'evaluation %i' % i)
        _verify_spec_tags(self.get('outputs', {}), ('type', 'value'),
                          'outputs')
        _verify_spec_tags(self.get('inputs', {}), ('type', 'value'),
                          'inputs')

    def __setitem__(self, key, value):
        if not key in __allowed_spec_keys__:
            _raise(ValueError, "refuse to add unsupported key", key)
        if key == 'version':
            if not isinstance(value, int) or value < 0:
                _raise(ValueError,
                    "version needs to be a non-negative integer value "
                    "(got: %s)." % value)
        dict.__setitem__(self, key, value)

    def get(self, *args):
        # check for proper field names
        if len(args) and not args[0] in __allowed_spec_keys__:
            raise ValueError("refuse to access unsupported key", args[0])
        return super(SPEC, self).get(*args)

    def get_hash(self):
        from hashlib import sha1
        str_repr = json.dumps(self, separators=(',',':'), sort_keys=True)
        return sha1(str_repr).hexdigest()

    def save(self, filename):
        spec_file = open(filename, 'w')
        spec_file.write(json.dumps(self, indent=2, sort_keys=True,
                                   cls=SPECJSONEncoder))
        spec_file.write('\n')
        spec_file.close()

    def _get_dict_specs(self, category, spec_type):
        if not category in self:
            return {}
        if spec_type is None:
            return self[category]
        else:
            specs = self[category] 
            return dict([(s, specs[s]) for s in specs if specs[s]['type'] == spec_type])

    def get_inputs(self, type_=None):
        return self._get_dict_specs('inputs', spec_type=type_)

    def get_outputs(self, type_=None):
        return self._get_dict_specs('outputs', spec_type=type_)

    def diff(self, spec, **kwargs):
        return diff(self, spec, **kwargs)


def spec_testoutput_ids(spec):
        return spec.get('outputs', {}).keys()

def diff(fr, to, recursiv_list=False):
    if not type(fr) == type(to):
        # different type
        return {'from': fr, 'to': to, '%%magic%%': 'diff'}
    elif fr is None and to is None:
        return None
    elif isinstance(fr, dict):
        dtree = {}
        # a dict
        fr_keys = set(fr.keys())
        to_keys = set(to.keys())
        # keys in fr but not in to
        for missing in fr_keys - to_keys:
            dtree[missing] = {'from': fr, '%%magic%%': 'diff'}
        # keys in to but not in fr
        for missing in to_keys - fr_keys:
            dtree[missing] = {'to': to, '%%magic%%': 'diff'}
        # compare intersecting keys
        for key in fr_keys.intersection(to_keys):
            value_diff = diff(fr[key], to[key], recursiv_list)
            if not value_diff is None:
                dtree[key] = value_diff
        if len(dtree):
            return dtree
        else:
            return None
    elif isinstance(fr, basestring):
        # any string
        if not fr == to:
            return {'ndiff': difflib.ndiff(('%s\n' % fr).splitlines(True),
                                           ('%s\n' % to).splitlines(True)),
                    '%%magic%%': 'diff'}
        else:
            return None
    elif isinstance(fr, float) or isinstance(fr, int):
        if not fr == to:
            return {'numdiff': to - fr, '%%magic%%': 'diff'}
        else:
            return None
    elif isinstance(fr, list):
        try:
            seqmatch = difflib.SequenceMatcher(None, fr, to).get_opcodes()
        except TypeError:
            raise NotImplementedError(
                "comparing sequences with unhashable values is not supported")
        if not len(seqmatch):
            return None
        elif len(seqmatch) == 1:
            # simple cases
            if seqmatch[0][0] == 'equal':
                # all the same
                return None
            if recursiv_list:
                if seqmatch[0][0] == 'replace':
                    # all different
                    return {'from': fr, 'to': to, '%%magic%%': 'diff'}
                else:
                    # either 'from' or 'to' were empty
                    return {'from': fr, 'to': to, '%%magic%%': 'diff'}
        if recursiv_list and \
            len([None for s in seqmatch
                 if s[0] == 'equal'
                    or (s[0] == 'replace' 
                          and s[2] == s[4] and s[1] == s[3])]) \
            == len(seqmatch):
            # only some values changed, but hopefully no shifts
            out = []
            for s in seqmatch:
                if s[0] == 'equal':
                    out.extend(fr[s[1]:s[2]])
                elif s[0] == 'replace':
                    for i in xrange(s[1], s[2]):
                        out.append(diff(fr[i], to[i], recursiv_list))
                else:
                    # all other conditions should be caught by top-level IF
                    raise RuntimeError('impossible opcode in sequence match')
            return out
        else:
            # complicated
            return {'seqmatch': seqmatch, '%%magic%%': 'diff'}
    raise RuntimeError('unhandled condition is SPEC diff')





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

import os
import re
import subprocess
import select
import datetime
import hashlib
import platform
from .pkg_mngr import PkgManager
import testkraut
from os.path import join as opj
from .spec import SPEC
import logging
lgr = logging.getLogger(__name__)

def which(program):
    """
    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    """
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    def ext_candidates(fpath):
        yield fpath
        for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
            yield fpath + ext

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            for candidate in ext_candidates(exe_file):
                if is_exe(candidate):
                    return candidate

    return None


class Stream(object):
    # this has been taken from Nipype (BSD-3-clause license)
    """Function to capture stdout and stderr streams with timestamps

    http://stackoverflow.com/questions/4984549/merge-and-sync-stdout-and-stderr/5188359#5188359
    """

    def __init__(self, name, impl):
        self._name = name
        self._impl = impl
        self._buf = ''
        self._rows = []
        self._lastidx = 0

    def fileno(self):
        "Pass-through for file descriptor."
        return self._impl.fileno()

    def read(self, drain=0):
        "Read from the file descriptor. If 'drain' set, read until EOF."
        while self._read(drain) is not None:
            if not drain:
                break

    def _read(self, drain):
        "Read from the file descriptor"
        fd = self.fileno()
        buf = os.read(fd, 4096)
        if not buf and not self._buf:
            return None
        if '\n' not in buf:
            if not drain:
                self._buf += buf
                return []

        # prepend any data previously read, then split into lines and format
        buf = self._buf + buf
        if '\n' in buf:
            tmp, rest = buf.rsplit('\n', 1)
        else:
            tmp = buf
            rest = None
        self._buf = rest
        now = datetime.datetime.now().isoformat()
        rows = tmp.split('\n')
        self._rows += [(now, '%s %s:%s' % (self._name, now, r), r) for r in rows]
        self._lastidx = len(self._rows)



def run_command(cmdline, cwd=None, env=None, timeout=0.01):
    # this has been taken from Nipype (BSD-3-clause license)
    """
    Run a command, read stdout and stderr, prefix with timestamp. The returned
    runtime contains a merged stdout+stderr log with timestamps

    http://stackoverflow.com/questions/4984549/merge-and-sync-stdout-and-stderr/5188359#5188359
    """
    PIPE = subprocess.PIPE
    proc = subprocess.Popen(cmdline,
                            stdout=PIPE,
                            stderr=PIPE,
                            shell=True,
                            cwd=cwd,
                            env=env)
    streams = [
        Stream('stdout', proc.stdout),
        Stream('stderr', proc.stderr)
        ]

    def _process(drain=0):
        try:
            res = select.select(streams, [], [], timeout)
        except select.error, e:
            if e[0] == errno.EINTR:
                return
            else:
                raise
        else:
            for stream in res[0]:
                stream.read(drain)

    while proc.returncode is None:
        proc.poll()
        _process()
    returncode = proc.returncode
    _process(drain=1)

    # collect results, merge and return
    result = {}
    temp = []
    for stream in streams:
        rows = stream._rows
        temp += rows
        result[stream._name] = [r[2] for r in rows]
    temp.sort()
    result['merged'] = [r[1] for r in temp]
    result['retval'] = returncode
    return result

def get_shlibdeps(binary):
    # for now only unix
    cmd = 'ldd %s' % binary
    ret = run_command(cmd)
    if not ret['retval'] == 0:
        raise RuntimeError("An error occurred while executing '%s'\n%s"
                           % (cmd, '\n'.join(ret['stderr'])))
    else:
        deps = [re.match(r'.*=> (.*) \(.*', l) for l in ret['stdout']]
        return [d.group(1) for d in deps if not d is None and len(d.group(1))]

def get_script_interpreter(filename):
    shebang = open(filename).readline()
    match = re.match(r'^#!(.*)$', shebang)
    if match is None:
        raise ValueError("no valid shebang line found in '%s'" % filename)
    return match.group(1).strip()

def hash(filename, method):
    hash = method
    with open(filename,'rb') as f: 
        for chunk in iter(lambda: f.read(128*hash.block_size), b''): 
             hash.update(chunk)
    return hash.hexdigest()

def sha1sum(filename):
    return hash(filename, hashlib.sha1())

def md5sum(filename):
    return hash(filename, hashlib.md5())

def _get_next_pid_id(procs, pid):
    base_pid = pid
    pid_suffix = 0
    while pid in procs:
        pid = '%s.%i' % (base_pid, pid_suffix)
        pid_suffix += 1
    return pid

def _find_parent_with_argv(procs, proc, match_argv):
    if proc['started_by'] is None:
        # reached the top
        return proc['pid']
    parent_proc = procs[proc['started_by']]
    if not parent_proc['argv'] is None \
       and not match_argv.match(parent_proc['argv'][0]) is None:
        return parent_proc['pid']
    else:
        return _find_parent_with_argv(procs, parent_proc, match_argv)

def _get_new_proc(procs, pid):
    oldpid = None
    if pid in procs:
        # archive a potentially existing proc of this PID under a safe
        # new PID
        oldpid = _get_next_pid_id(procs, pid)
        proc = procs[pid]
        proc['pid'] = oldpid
        procs[oldpid] = proc
    proc = dict(dict(pid=pid, started_by=None, argv=None,
                     uses=[], generates=[]))
    procs[pid] = proc
    return proc, oldpid

def get_cmd_prov_strace(cmd, match_argv=None):
    if match_argv is None:
        match_argv = r'.*'
    match_argv = re.compile(match_argv)
    cmd_prefix = ['strace', '-q', '-f', '-s', '1024',
                  '-e', 'trace=execve,clone,open,openat,unlink,unlinkat']
    cmd = cmd_prefix + cmd
    cmd_exec = subprocess.Popen(cmd, bufsize=0,
                             stderr=subprocess.PIPE)
    # store discovered processes
    procs = {}
    # store accessed files
    files = {}
    curr_proc = None
    # precompile REs
    quoted_list_splitter = re.compile(r'(?:[^,"]|"[^"]*\")+')
    syscall_arg_splitter = re.compile(r'(?:[^,[]|\[[^]]*\])+')
    #strace_ouput_splitter = re.compile(r'^(\[pid\s+([0-9]+)\] |)([a-z0-9_]+)\((.*)\) (.*)')
    strace_output_splitter = re.compile(r'^(\[pid\s+([0-9]+)\] |)([a-z0-9_]+)\((.*)')
    strace_resume_splitter = re.compile(r'^(\[pid\s+([0-9]+)\] |)<\.\.\. ([a-z0-9_]+) resumed> (.*)')
    unfinished_splitter = re.compile(r'(.*)\s+<unfinished \.\.\.>')
    rest_splitter = re.compile(r'(.*)\s+=\s+(.*)')
    # for every line in strace's output
    root_pid = None
    unfinished = {}
    for line in cmd_exec.stderr:
        match = strace_output_splitter.match(line)
        if match is None:
            # this could be a resume line
            match = strace_resume_splitter.match(line)
            if match is None:
                # ignore funny line
                continue
            # we have a resume, check if we know the beginning of it
            _, pid, syscall, rest = match.groups()
            if pid in unfinished:
                pdict = unfinished[pid]
                if syscall in pdict:
                    start = pdict[syscall]
                    del pdict[syscall]
                else:
                    raise RuntimeError("no resume info on started syscall (%s, %s)"
                                       % (syscall, pid))
                if not len(pdict):
                    del unfinished[pid]
            else:
                raise RuntimeError("no resume info on pid %s"
                                   % pid)
            rest = '%s %s' % (start, rest)
        else:
            #_, pid, syscall, syscall_args, syscall_ret = match.groups()
            _, pid, syscall, rest = match.groups()
            umatch = unfinished_splitter.match(rest)
            if not umatch is None:
                # this is the start of an unfinished syscall
                pdict = unfinished.get(pid, dict())
                pdict[syscall] = umatch.group(1)
                unfinished[pid] = pdict
                continue # will be processed on resume
        syscall_args, syscall_ret = rest_splitter.match(rest).groups()
        if not pid is None and not pid == root_pid and not pid in procs:
            if not root_pid is None:
                raise RuntimeError("we already have a root PID, and found a new one")
            root_pid = pid
        if pid is None:
            pid = 'mother'
        if syscall_ret.startswith('-'):
            # ignore any syscall that yielded an error
            continue
        # everything we know about this process
        if not pid in procs:
            proc, _ = _get_new_proc(procs, pid)
        else:
            proc = procs[pid]
        # split the syscall args into a list
        syscall_args = syscall_arg_splitter.findall(syscall_args)
        if syscall == 'clone':
            newpid = syscall_ret
            # it started a new proc
            new_proc, _ = _get_new_proc(procs, newpid)
            new_proc['started_by'] = pid
        elif syscall == 'execve':
            # start a process
            executable = syscall_args[0].strip('"')
            argv = [arg.strip(' "') for arg in
                        quoted_list_splitter.findall(syscall_args[1].strip(' []'))]
            if not proc['argv'] is None:
                # a new command in the same process -> code as a new process)
                new_proc, oldpid = _get_new_proc(procs, pid)
                new_proc['started_by'] = oldpid
                proc = new_proc
            proc.update(dict(executable=executable,
                             argv=argv))
        elif syscall == 'open':
            # open a file
            open_args = [arg.strip(' "') for arg in syscall_args]
            filename = os.path.relpath(open_args[0])
            access_mode = open_args[1]
            if filename.startswith(os.path.pardir):
                # track files under the current dir only
                continue
            if 'O_WRONLY' in access_mode or 'O_RDWR' in access_mode:
                proc['generates'].append(filename)
            elif 'O_RDONLY' in access_mode or 'O_RDWR' in access_mode:
                proc['uses'].append(filename)
        else:
            # ignore all other syscalls
            pass
    # rewrite PID of the root process if we got to know it
    if not root_pid is None:
        # merge the info of root_pid with the mother's
        rproc = procs[root_pid]
        for pid, proc in procs.iteritems():
            if pid.startswith('mother'):
                for attr in ('generates', 'uses'):
                    proc[attr] += rproc[attr]
                    proc[attr] = [a.replace('mother', root_pid)
                                    for a in proc[attr]]
                proc['pid'] = pid.replace('mother', root_pid)
            for attr in ('started_by',):
                if not proc[attr] is None:
                    proc[attr] = proc[attr].replace('mother', root_pid)
            #REPLACE ALL MOTHER REFERENCES IN ALL ATTRS
        del procs[root_pid]
        procs = dict([(pid.replace('mother', root_pid), info) for pid, info in procs.iteritems()])
    # uniquify
    for pid, proc in procs.iteritems():
        for attr in ('generates', 'uses'):
            proc[attr] = set(proc[attr])
    # rewrite inter-proc dependencies to point to processes with cmdinfo
    pid_mapper = {}
    for pid, proc in procs.iteritems():
        if proc['started_by'] is None:
            # cache pid of the mother
            if 'argv' in proc:
                pid_mapper[pid] = pid
            # nothing else to recode
            continue
        # we have a parent process, but we might have no cmd info
        # -> retrace graph upwards to find a parent with info
        parent_pid = proc['started_by']
        new_parent_pid = pid_mapper.get(parent_pid,
                                        _find_parent_with_argv(procs,
                                                               proc,
                                                               match_argv))
        # cache pid mapping: old parent -> new parent
        # (even if it woudl be the same)
        pid_mapper[parent_pid] = new_parent_pid
        # rewrite parent in current process
        proc['started_by'] = new_parent_pid
        if proc['argv'] is None \
           or match_argv.match(['argv'][0]) is None:
            # we don't want to know about this process
            # move the files it uses and generates upwards
            new_parent_proc = procs[new_parent_pid]
            for field in ('uses', 'generates'):
                new_parent_proc[field] = new_parent_proc[field].union(proc[field])
        else:
            # this proc info stays
            pid_mapper[pid] = pid
    # filter all procs that have no argv
    procs = dict([(pid, procs[pid]) for pid in pid_mapper.values()])
    # wait() sets the returncode
    cmd_exec.wait()
    return procs, cmd_exec.returncode

def guess_file_tags(fname):
    tags = set()
    if not os.path.getsize(fname):
        # no futher tags for empty files
        tags.add('empty')
        return tags
    try:
        import nibabel as nb
        img = nb.load(fname)
        tags.add('volumetric image')
        tags.add('%iD image' % len(img.get_shape()))
    except:
        pass
    try:
        import numpy as np
        mat = np.loadtxt(fname)
        tags.add('space-separated values')
        tags.add('text file')
    except:
        pass
    return tags

def describe_system():
    sysinfo = {}
    for fx in ('architecture', 'machine', 'python_build', 'python_compiler',
               'python_branch', 'python_implementation', 'python_revision',
               'python_version', 'release', 'system'):
        if hasattr(platform, fx):
            sysinfo[fx] = getattr(platform, fx)()
    # core deps
    for pkg in ('numpy', 'scipy', 'nibabel'):
        try:
            imp_pkg = __import__(pkg, globals(), locals(), [], -1)
            sysinfo['pkg_%s_version' % pkg] = imp_pkg.__version__
        except (ImportError, AttributeError):
            sysinfo['pkg_%s_version' % pkg] = None
    # trigger platform hooks
    if len(sysinfo.get('system', '')):
        system_descr = '_describe_%s_system' % sysinfo['system'].lower()
        if system_descr in globals():
            globals()[system_descr](sysinfo)
    return sysinfo

def _describe_linux_system(sysinfo):
    for fx in ('linux_distribution', 'dist'):
        if hasattr(platform, fx):
            sysinfo[fx] = getattr(platform, fx)()
    if 'dist' in sysinfo:
        # only keep the modern variant
        if not 'linux_distribution' in sysinfo:
            sysinfo['linux_distribution'] = sysinfo['dist']
        del sysinfo['dist']
    # system hooks
    distr_info = sysinfo.get('linux_distribution', '')
    if len(distr_info):
        distr_descr = '_describe_%s_system' % distr_info[0].lower()
        if distr_descr in globals():
            globals()[distr_descr](sysinfo)

def _describe_darwin_system(sysinfo):
    for fx in ('mac_ver'):
        if hasattr(platform, fx):
            sysinfo[fx] = getattr(platform, fx)()

def _describe_debian_system(sysinfo):
    pkg_mngr = PkgManager()
    for pkg in ('numpy', 'scipy', 'nibabel', 'apt'):
        sysinfo['pkg_%s' % pkg] = pkg_mngr.get_pkg_info('python-%s' % pkg)

def _describe_ubuntu_system(sysinfo):
    _describe_debian_system(sysinfo)

def get_test_library_paths(prepend=None):
    """Returns a sequence with all configured tets library paths.

    Parameters
    ==========
    prepend : list
      sequence with additional paths that are prepended to the output
    """
    testlibdirs = [os.path.expandvars(tld) for tld in
                      testkraut.cfg.get('library', 'paths',
                              default=opj(os.curdir, 'library')).split(',')]
    # always add the built-in library as a last resort
    testlibdirs.append(opj(os.path.dirname(testkraut.__file__), 'library'))
    if not prepend is None:
        return prepend + testlibdirs
    else:
        return testlibdirs

def get_spec(spec_def, libraries=None):
    """Return a SPEC object from a number of obscure locations.

    SPEC can be given as a string, a path to a SPEC file, or a SPEC
    ID that is search for in any configured library location (plus
    additional libraries passed via ``libraries``).
    """
    spec = None
    # look for the SPEC in any possible library
    for tld in get_test_library_paths(libraries):
        testlib_filepath = opj(tld, spec_def, 'spec.json')
        if os.path.isfile(testlib_filepath):
            lgr.debug("located SPEC for test '%s' in library at '%s'"
                      % (spec_def, tld))
            spec = SPEC(open(testlib_filepath))
            break
        else:
            lgr.debug("did not find SPEC for test '%s' in library at '%s'"
                      % (spec_def, tld))
    if spec is None and os.path.isfile(spec_def):
        # open explicit spec file
        spec = SPEC(open(spec_def))
    if spec is None:
        # spec is given as a str?
        try:
            spec = SPEC(spec_def)
        except ValueError:
            # not a SPEC
            raise ValueError("'%s' is neither a SPEC, or a " % spec_def +
                             "filename of an existing SPEC file, or an ID "
                             "of a test in any of the configured libraries.")
    return spec



import os
import nipype.interfaces.fsl as fsl          # fsl
import nipype.interfaces.io as nio           # Data i/o
import nipype.pipeline.engine as pe          # pypeline engine

fsl.FSLCommand.set_default_output_type('NIFTI_GZ')

test_workflow = pe.Workflow(name='justbet')

datasource = pe.Node(interface=nio.DataGrabber(outfields=['head']),
                     name = 'datasource')
datasource.inputs.base_directory = os.path.abspath(os.curdir)
datasource.inputs.template = '%s.nii.gz'
datasource.inputs.template_args = dict(head=[['head']])


bet = pe.Node(interface=fsl.BET(mask=True),
              name='bet')

datasink = pe.Node(interface=nio.DataSink(parameterization=False),
                   name="datasink")
datasink.inputs.base_directory = os.path.abspath(os.curdir)
test_workflow.connect(datasource, 'head', bet, 'in_file')
test_workflow.connect(bet, 'out_file', datasink, 'output.@brain')
test_workflow.connect(bet, 'mask_file', datasink, 'output.@brainmask')



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

import logging
lgr = logging.getLogger(__name__)
import os

class PkgManager(object):
    """Simple abstraction layer to query local package managers"""
    def __init__(self):
        self._mode = None
        self._native_pkg_cache = None
        from .utils import run_command
        try:
            import apt
            self._native_pkg_cache = apt.Cache()
            self._mode = 'deb'
        except ImportError:
            # it could still be debian, but without python-apt
            ret = run_command('dpkg --version')
            if ret['retval'] == 0:
                self._mode = 'deb'
                lgr.warning("Running on Debian platform but no python-apt "
                            "package found -- only limited information is "
                            "available.")
        # maybe RPM?
        ret = run_command('rpm --version')
        if ret['retval'] == 0:
            self._mode = 'rpm'

    def get_pkg_name(self, filename):
        """Return the name of a package providing a file (if any).

        Returns None if no package provides this file, or no package manager is
        available.
        """
        if os.path.exists(filename):
            # if the file actually exists try resolving symlinks
            filename = os.path.realpath(filename)
        if self._mode == 'deb':
            return _get_debian_pkgname(filename)
        elif self._mode == 'rpm':
            from .utils import run_command
            ret = run_command("rpm --queryformat '%%{NAME}\n' -qf %s" % filename)
            if not ret['retval'] == 0:
                return None
            return ret['stdout'][0]
        return None

    def get_pkg_info(self, pkgname):
        """Returns a dict with information on a given package."""
        info = dict(name=pkgname)
        if self._mode == 'deb':
            return self._get_debian_pkginfo(pkgname, info)
        elif self._mode == 'rpm':
            from .utils import run_command
            ret = run_command('rpm --qf \'\{"version":"%%{EVR}", "sha1sum":"%%{SHA1HEADER}", "vendor":"%%{VENDOR}", "arch":"%%{ARCH}"\}\n\' -q %s' % pkgname)
            if ret['retval'] == 0:
                info.update(eval('\n'.join(ret['stdout'])))
        return info

    def _get_debian_pkginfo(self, pkgname, debinfo):
        apt = self._native_pkg_cache
        if not apt is None:
            pkg = apt[pkgname].installed
            if pkg is None:
                # no such package installed
                return debinfo
            debinfo['version'] = pkg.version
            debinfo['sha1sum'] = pkg.sha1
            debinfo['arch'] = pkg.architecture
            origin = pkg.origins[0]
            debinfo['vendor'] = origin.origin
        return debinfo

    def get_platform_name(self):
        """Returns the local package manager type.

        For example, 'deb', 'rpm', or None if no supported package manager
        was found.
        """
        return self._mode


def _get_debian_pkgname(filename):
    from .utils import run_command
    # provided by a Debian package?
    pkgname = None
    try:
        ret = run_command('dpkg -S %s' % filename)
    except OSError:
        return None
    if not ret['retval'] == 0:
        return None
    for line in ret['stdout']:
        lspl = line.split(':')
        if lspl[0].count(' '):
            continue
        pkgname = lspl[0]
        break
    return pkgname



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

from testkraut.fingerprinting import get_fingerprinters
from nose.tools import *

def test_fingerprint_dict():
    vol_fp = get_fingerprinters('volumetric image')
    assert_true(len(vol_fp))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

import operator
import logging
lgr = logging.getLogger(__name__)

_tag2fx = {}

def get_fingerprinters(tag):
    if not len(_tag2fx):
        # charge
        from testkraut import cfg
        tags = set(cfg.options('system fingerprints')).union(cfg.options('fingerprints'))
        for tag in tags:
            fp_tag = set()
            for fps_str in cfg.get('system fingerprints', tag, default="").split() \
                         + cfg.get('fingerprints', tag, default="").split():
                fps_comp = fps_str.split('.')
                try:
                    mod = __import__('.'.join(fps_comp[:-1]), globals(), locals(),
                                     fps_comp[-1:], -1)
                    fps = getattr(mod, fps_comp[-1])
                except:
                    lgr.warning(
                        "ignoring invalid fingerprinting function '%s' for tag '%s'"
                        % (fps_str, tag))
                fp_tag.add(fps)
            _tag2fx[tag] = fp_tag
    return _tag2fx.get(tag, set())

def fp_volume_image(fname, fp):
    # this version needs an increment whenever this implementation changes
    fp['version'] = 0
    import nibabel as nb
    import numpy as np
    from scipy.ndimage import measurements as msr
    from scipy.stats import describe
    img = nb.load(fname)
    img_data = img.get_data().astype('float') # float for z-score
    # keep a map where the original data is larger than zero
    zero_thresh = img_data > 0
    # basic descriptive stats
    img_size, img_minmax, img_mean, img_var, img_skew, img_kurt = \
            describe(img_data, axis=None)
    img_std = np.sqrt(img_var)
    fp['std'] = img_std
    fp['mean'] = img_mean
    fp['min'] = img_minmax[0]
    fp['max'] = img_minmax[1]
    fp['skewness'] = img_skew
    fp['kurtosis'] = img_kurt
    # global zscore 
    img_data -= img_mean
    img_data /= img_std
    zmin = img_data.min()
    zmax = img_data.max()
    # normalized luminance histogram (needs zscores)
    luminance_hist_params = (-3, 3, 7)
    fp['luminance_histogram_[%i,%i,%i]' % luminance_hist_params] = \
            np.histogram(img_data, normed=True,
                         bins=np.linspace(*luminance_hist_params))[0]
    if not img_std:
        # no variance, nothing to do
        return
    # perform thresholding at various levels and compute descriptive
    # stats of the resulting clusters
    clusters = np.empty(img_data.shape, dtype=np.int32)
    for thresh in (zero_thresh, 2.0, 4.0, 8.0, -2.0, -4.0, -8.0):
        # create a binarized map for the respective threshold
        if not isinstance(thresh, float):
            thresh_map = thresh
            thresh = 'orig_zero'
        else:
            if thresh < 0:
                if thresh < zmin:
                    # thresholding would yield nothing
                    continue
                thresh_map = img_data < thresh
            else:
                if thresh > zmax:
                    # thresholding would yield nothing
                    continue
                thresh_map = img_data > thresh
        nclusters = msr.label(thresh_map, output=clusters)
        # sort by cluster size
        cluster_sizes = [(cl, np.sum(clusters == cl))
                                for cl in xrange(1, nclusters + 1)]
        cluster_sizes = sorted(cluster_sizes,
                               key=operator.itemgetter(1),
                               reverse=True)
        # how many clusters to report
        max_nclusters = 3
        if not len(cluster_sizes):
            # nothing to report, do not clutter the dict
            continue
        clinfo = {}
        fp['thresh_%s' % thresh] = clinfo
        clinfo['nclusters'] = nclusters
        # only for the biggest clusters
        cl_id = 0
        for cl_label, cl_size in cluster_sizes[:max_nclusters]:
            cl_id += 1
            cli = dict(size=cl_size)
            clinfo['cluster_%i' % cl_id] = cli
            # center of mass of the cluster extent (ignoring actual values)
            cli['extent_ctr_of_mass'] = \
                    msr.center_of_mass(thresh_map, labels=clusters,
                                       index=cl_label)
            # center of mass of the cluster considering actual values
            cli['ctr_of_mass'] = msr.center_of_mass(img_data,
                                                    labels=clusters,
                                                    index=cl_label)
            if isinstance(thresh, float) and thresh < 0:
                # position of minima
                pos = msr.minimum_position(img_data, labels=clusters, index=cl_label)
                cli['min_pos'] = pos
                cli['min'] = img_data[pos]
            else:
                # position of maxima
                pos = msr.maximum_position(img_data, labels=clusters, index=cl_label)
                cli['max_pos'] = pos
                cli['max'] = img_data[pos]

def fp_nifti(fname, fp):
    pass



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

__version__ = '0.0~pre2'

import os
from .config import ConfigManager

class _SingletonType(type):
    """Simple singleton implementation adjusted from
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/412551
    """
    def __init__(mcs, *args):
        type.__init__(mcs, *args)
        mcs._instances = {}

    def __call__(mcs, sid, instance, *args):
        if not sid in mcs._instances:
            mcs._instances[sid] = instance
        return mcs._instances[sid]

class __Singleton:
    """To ensure single instance of a class instantiation (object)

    """

    __metaclass__ = _SingletonType
    def __init__(self, *args):
        pass
    # Provided __call__ just to make silly pylint happy
    def __call__(self):
        raise NotImplementedError

#
# As the very first step: Setup configuration registry instance and
# read all configuration settings from files and env variables
#
_cfgfile = os.environ.get('TESTKRAUTCONFIG', None)
if _cfgfile:
    # We have to provide a list
    _cfgfile = [_cfgfile]
cfg = __Singleton('cfg', ConfigManager(_cfgfile))





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Generate a test specification from an arbitrary command call.

some more bits...

Doesn't work nicely when input data is under /tmp

Examples:

$ testkraut generate --id demo -o test.spec -- bash -c "cat spec.json > newstuff"

TODO: ADD OUTPUT DESCRIPTION
"""

__docformat__ = 'restructuredtext'

# magic line for manpage summary
# man: -*- % generate a test SPEC from an arbitrary command call

import argparse
import os
import re
import itertools
from os.path import join as opj
from ..spec import SPEC
from ..utils import sha1sum, get_cmd_prov_strace, guess_file_tags
from ..pkg_mngr import PkgManager

parser_args = dict(formatter_class=argparse.RawDescriptionHelpFormatter)

def setup_parser(parser):
    parser.add_argument(
        '-o', '--spec-file', required=True, metavar='FILENAME', dest='spec_filename',
        help="name of the output SPEC file")
    parser.add_argument(
        '--id', required=True,
        help="SPEC name/identifier")
    parser.add_argument(
        '--description', default='',
        help="SPEC description")
    parser.add_argument(
        '--author', default='', nargs=2,
        help="SPEC author given as <'NAMES' EMAIL>")
    parser.add_argument(
        '--sv', '--spec-version', default=0, type=int, dest='spec_version',
        metavar='VERSION', help="SPEC version")
    parser.add_argument(
        '--env', '--dump-environment', metavar='REGEX', dest='dump_env',
        help="""dump all environment variables into the SPEC whose names match
             the regular expression""")
    parser.add_argument(
        '--no-strace', action='store_true',
        help="do not use strace to analyze software and data dependencies.")
    parser.add_argument(
        '--ignore-outputs',
        help="regular expression matching command output filenames/paths to ignore.")
    parser.add_argument(
        '--match-outputs',
        help="""regular expression matching command output filenames/paths to
             consider in the SPEC -- ignore all others.""")
    parser.add_argument(
        '--match-cmds', default=r'.*', metavar='REGEX',
        help="regular expression matching commands to capture as test components.")
    parser.add_argument(
        '--nomin', '--no-minimize-inputs', action='store_true', dest='no_minimize_inputs',
        help="""always include all files present in the test directory as input
             files, regardless of whether they are actually used during
             execution. This option only has an effect when strace is used. If it
             is not available, no minimization is performed anyway. If strace is
             not able to properly track file access, resulting in insufficent
             input SPECs, enable the option.""")
    parser.add_argument('arg', nargs='+', metavar='ARGS',
        help="""command or workflow filename""")

def get_dir_hashes(path, ignore=None):
    if ignore is None:
        ignore = []
    dir_content = []
    for dirlist in os.walk(path):
        tocheck = [fn for fn in dirlist[2] if not fn in ignore]
        nfiles = len(tocheck)
        dirfiles = [fn for fn in itertools.imap(opj,
                                                [dirlist[0]] * nfiles,
                                                tocheck)]
        dir_content += [(fn, sha1sum(fn)) for fn in dirfiles if os.path.isfile(fn)]
    return dict(dir_content)

def find_executables(path):
    executables = []
    for dirlist in os.walk(path):
        for fn in dirlist[2]:
            testfile = opj(dirlist[0], fn)
            if os.path.isfile(testfile) and os.access(testfile, os.X_OK):
                executables.append(testfile)
    return executables

def find_files(path):
    files = []
    for dirlist in os.walk(path):
        for fn in dirlist[2]:
            testfile = opj(dirlist[0], fn)
            if os.path.isfile(testfile):
                files.append(testfile)
    return files

def _proc_generates(procs, proc, starts, filter=None, ignore=None):
    generated = proc.get('generates', [])
    if len(generated):
        if not ignore is None:
            generated = [g for g in generated if ignore.match(g) is None]
        if not filter is None:
            generated = [g for g in generated if not filter.match(g) is None]
        if len(generated):
            # still files left?
            return True
    for started_pid in starts[proc['pid']]:
        generates = _proc_generates(procs, procs[started_pid], starts,
                                    filter=filter, ignore=ignore)
        if generates:
            return True
    return False

def run(args):
    import sys
    import subprocess
    import tempfile
    import re
    import shutil
    import json

    # try using a local package manager to obtain more info on software deps
    pkg_mngr = PkgManager()

    # assume execution within the testbed
    testbed_dir = os.path.abspath(os.curdir)
    # get the state of the union
    prior_test_hashes = get_dir_hashes(testbed_dir)
    # run through strace
    proc_info, retval = get_cmd_prov_strace(args.arg, args.match_cmds)
    if not retval == 0:
        raise RuntimeError('command returned with non-zero exit code %s'
                           % args.arg)
    used_files = set()
    starts = dict(zip([p['pid'] for p in proc_info.values()],
                      [list() for i in xrange(len(proc_info))]))
    for proc in proc_info.values():
        used_files = used_files.union(proc['uses'])
        if not proc['started_by'] is None:
            starts[proc['started_by']].append(proc['pid'])
    # testbed content after run
    post_test_hashes = get_dir_hashes(testbed_dir)
    # categorize testbed content
    new_files = set(post_test_hashes).difference(prior_test_hashes)
    deleted_files = set(prior_test_hashes).difference(post_test_hashes)
    changed_files = [fn for fn in prior_test_hashes if not fn in deleted_files and prior_test_hashes[fn] != post_test_hashes[fn]]

    # spec skeleton
    spec = SPEC(
            dict(id=args.id,
                description=args.description,
                version=args.spec_version,
                processes={},
                executables={},
                environment={},
                test={},
                inputs={},
                outputs={},
                evaluations={}))

    # in case of a shell command
    spec['test'] = dict(type='shell_command', command=args.arg)
    # filter files
    if not args.ignore_outputs is None:
        args.ignore_outputs = re.compile(args.ignore_outputs)
    if not args.match_outputs is None:
        args.match_outputs = re.compile(args.match_outputs)
    # record al input files
    for ipf in prior_test_hashes:
        relname = os.path.relpath(ipf)
        if not relname in used_files:
            # skip
            continue
        s = dict(type='file', value=relname,
                 sha1sum=prior_test_hashes[ipf],
                 tags=list(guess_file_tags(relname)))
        spec['inputs'][relname] = s
    # record all output files
    for opf in new_files:
        relname = os.path.relpath(opf)
        if (not args.ignore_outputs is None and not args.ignore_outputs.match(relname) is None) \
           or (not args.match_outputs is None and args.match_outputs.match(relname) is None):
            # skip
            continue
        s = dict(type='file', value=relname,
                 tags=list(guess_file_tags(relname)))
        spec['outputs'][relname] = s

    # and now get all info into the SPEC
    dep_mapper = {}
    spec['dependencies'] = dep_mapper
    proc_mapper = {}
    spec['processes'] = proc_mapper
    exec_mapper = {}
    spec['executables'] = exec_mapper
    pid_counter = 0
    pid_native2new = {}
    exec2pkg = set()
    # which processes are (indirectly) involved in some sort of output generation
    effective_pids = [proc['pid'] for pid, proc in proc_info.iteritems()
                        if _proc_generates(proc_info, proc, starts,
                           ignore=args.ignore_outputs,
                           filter=args.match_outputs)]
    for pid, proc in proc_info.iteritems():
        if not pid in effective_pids:
            # skip all procs that do not generate or start something that generates
            continue
        # discover the corresponding executable
        executable = os.path.realpath(proc['executable'])
        espec = dict()
        exec_mapper[executable] = espec
        # make a record of this process
        pspec = dict(executable=executable,
                     #pid=pid_counter, argv=proc['argv'])
                     argv=proc['argv'])
        for ftype in ('uses', 'generates'):
            flist = proc.get(ftype, [])
            if len(flist):
                pspec[ftype] = [os.path.relpath(fn) for fn in flist if os.path.isfile(fn)]
        if 'started_by' in proc:
            pspec['started_by'] = proc['started_by']
        proc_mapper[pid_counter] = pspec
        ## map old to new PID
        pid_native2new[proc['pid']] = pid_counter
        pid_counter += 1
        if not executable in exec2pkg:
            # we haven't seen this binary yet
            pkgname = pkg_mngr.get_pkg_name(executable)
            if not pkgname is None:
                pkg_deps = dep_mapper.get(pkg_mngr.get_platform_name(), set())
                pkg_deps.add(pkgname)
                dep_mapper[pkg_mngr.get_platform_name()] = pkg_deps
            exec2pkg.add(executable)
    ## 2nd pass -- store inter-process deps using new/simplified PIDs
    for pid, proc in proc_mapper.iteritems():
        if not 'started_by' in proc:
            continue
        native_parent_pid = proc['started_by']
        if native_parent_pid is None:
            # this is the mother process
            continue
        proc['started_by'] = pid_native2new[native_parent_pid]
    if 'deb' in dep_mapper:
        # do a dpkg style dependency statement
        dep_mapper['deb'] = ' ,'.join(dep_mapper['deb'])
    # record full environment (if desired)
    if not args.dump_env is None:
        for env in os.environ:
            if not re.match(args.dump_env, env) is None:
                spec['environment'][env] = {}
    spec.save(args.spec_filename)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Execute a SPEC.

some more bits...

Examples:

$ testkraut execute demo.json

TODO: ADD OUTPUT DESCRIPTION
"""

__docformat__ = 'restructuredtext'

# magic line for manpage summary
# man: -*- % generate a test SPEC from an arbitrary command call

import argparse
import os
import sys
from os.path import join as opj
from ..spec import SPEC
from ..utils import get_spec
import testkraut
from testkraut import cfg

parser_args = dict(formatter_class=argparse.RawDescriptionHelpFormatter)

def setup_parser(parser):
    parser.add_argument('spec', metavar='SPEC',
        help="""SPEC filename or testlibary ID""")
    parser.add_argument('-l', '--library', action='append',
            help="""alternative path to a test library. When given, this setting
                 overwrites any library configuration setting""")

def run(args):
    lgr = args.logger
    from .. import runner as tkr
    from ..spec import SPECJSONEncoder
    runner = tkr.LocalRunner(testlibs=args.library)
    # obtain the full SPEC from magic storage
    spec = get_spec(args.spec, args.library)
    try:
        retval = runner(spec)
    finally:
        tbd = runner.get_testbed_dir(spec)
        if os.path.exists(tbd):
            # if we got a testbed, be sure to dump all info that we gathered
            spec.save(opj(tbd, 'spec.json'))
    if not retval:
        lgr.critical("test '%s' failed" % args.spec)
        lgr.info(spec.get('test', {}).get('stdout', ''))
        lgr.info(spec.get('test', {}).get('stderr', ''))
        raise RuntimeError("abort due to test failure")



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Determine the difference between to SPECs.

"""

__docformat__ = 'restructuredtext'

# magic line for manpage summary
# man: -*- % generate a test SPEC from an arbitrary command call

import sys
import argparse
from ..spec import SPEC

try:
    if not sys.stdout.isatty():
        raise Exception
    from colorama import init
    init(autoreset=True)
    from colorama import Fore as colors
    from colorama import Style as style
except:
    class DummyColors(object):
        def __getattribute__(self, name):
            return ''
    colors = DummyColors()
    style = DummyColors()

parser_args = dict(formatter_class=argparse.RawDescriptionHelpFormatter)

def setup_parser(parser):
    parser.add_argument('specs', nargs=2, metavar='SPEC',
            help="SPEC name/identifier")

def print_diff_hdr(fr, to, mode):
    print '%sdiff --%s\n--- %s\n+++ %s' % (style.BRIGHT, mode, fr, to)

def print_diff(breadcrumbs, diffspec, fr, to):
    bc_str = '->'.join(breadcrumbs)
    if 'ndiff' in diffspec:
        print_diff_hdr(bc_str, bc_str, 'ndiff')
        for ds in diffspec['ndiff']:
            if ds.startswith('-'):
                print colors.RED + ds,
            elif ds.startswith('+'):
                print colors.GREEN + ds,
            else:
                print colors.CYAN + ds,
    elif 'numdiff' in diffspec:
        print_diff_hdr(bc_str, bc_str, 'num')
        print '$ %s' % diffspec['numdiff']
    elif 'seqmatch' in diffspec:
        print_diff_hdr(bc_str, bc_str, 'seq')
        for oc in diffspec['seqmatch']:
            if oc[0] == 'equal':
                continue
            print colors.CYAN + '@@ -%i,%i +%i,%i @@' % (oc[1], oc[2] - oc[1],
                                                         oc[3], oc[4] - oc[3])
            if oc[0] in ('delete', 'replace'):
                for el in fr[oc[1]:oc[2]]:
                    print colors.RED + '- %s' % el
            if oc[0] in ('insert', 'replace'):
                for el in to[oc[3]:oc[4]]:
                    print colors.GREEN + '+ %s' % el
    else:
        frs = tos = bc_str
        if fr is None:
            frs = '(none)'
        if to is None:
            tos = '(none)'
        print_diff_hdr(frs, tos, 'spec')
        if not fr is None:
            print colors.RED + '- %s' % fr
        if not to is None:
            print colors.GREEN + '+ %s' % to

def walk_difftree(dt, fr, to, render, breadcrumbs=None):
    if breadcrumbs is None:
        breadcrumbs = []
    for i, k in enumerate(dt):
        if isinstance(dt, dict):
            v = dt[k]
            f = fr.get(k, None)
            t = to.get(k, None)
            breadcrumb = k
        else:
            v = k
            f = fr[i]
            t = to[i]
            breadcrumb = '(%i)' % i
        if not (isinstance(v, dict) or isinstance(v, list)):
            # non-container
            continue
        if isinstance(v, dict) and '%%magic%%' in v and v['%%magic%%'] == 'diff':
            # we found a diff -- render it
            render(breadcrumbs + [breadcrumb], v, f, t)
        else:
            # walk further
            walk_difftree(v, f, t, render, breadcrumbs + [breadcrumb])



def run(args):
    fspec = SPEC(open(args.specs[0]))
    tspec = SPEC(open(args.specs[1]))
    difftree = fspec.diff(tspec)
    walk_difftree(difftree, fspec, tspec, print_diff)



'''Python implemetation of the W3C Provenance Data Model (PROV-DM)

Support for PROV-JSON import/export 

References:

PROV-DM: http://www.w3.org/TR/prov-dm/

@author: Trung Dong Huynh <trungdong@donggiang.com>
@copyright: University of Southampton 2012
'''

import datetime
import json
import re
import collections
from collections import defaultdict
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
# XXX: temp_disable from .. import logging
# XXX: temp_disable logger = logging.getLogger('interface')

## PROV record constants - PROV-DM LC
# C1. Entities/Activities
PROV_REC_ENTITY                 = 1
PROV_REC_ACTIVITY               = 2
PROV_REC_GENERATION             = 11
PROV_REC_USAGE                  = 12
PROV_REC_COMMUNICATION          = 13
PROV_REC_START                  = 14
PROV_REC_END                    = 15
PROV_REC_INVALIDATION           = 16

# C2. Derivations
PROV_REC_DERIVATION             = 21

# C3. Agents/Responsibility
PROV_REC_AGENT                  = 3
PROV_REC_ATTRIBUTION            = 31
PROV_REC_ASSOCIATION            = 32
PROV_REC_DELEGATION             = 33
PROV_REC_INFLUENCE              = 34
# C4. Bundles
PROV_REC_BUNDLE                 = 4 # This is the lowest value, so bundle(s) in JSON will be decoded first 
# C5. Alternate
PROV_REC_ALTERNATE              = 51
PROV_REC_SPECIALIZATION         = 52
PROV_REC_MENTION                = 53
# C6. Collections
PROV_REC_COLLECTION             = 6
PROV_REC_MEMBERSHIP             = 61

PROV_RECORD_TYPES = (
    (PROV_REC_ENTITY,               u'Entity'),
    (PROV_REC_ACTIVITY,             u'Activity'),
    (PROV_REC_GENERATION,           u'Generation'),
    (PROV_REC_USAGE,                u'Usage'),
    (PROV_REC_COMMUNICATION,        u'Communication'),
    (PROV_REC_START,                u'Start'),
    (PROV_REC_END,                  u'End'),
    (PROV_REC_INVALIDATION,         u'Invalidation'),
    (PROV_REC_DERIVATION,           u'Derivation'),
    (PROV_REC_AGENT,                u'Agent'),
    (PROV_REC_ATTRIBUTION,          u'Attribution'),
    (PROV_REC_ASSOCIATION,          u'Association'),
    (PROV_REC_DELEGATION,           u'Delegation'),
    (PROV_REC_INFLUENCE,            u'Influence'),
    (PROV_REC_BUNDLE,               u'Bundle'),
    (PROV_REC_ALTERNATE,            u'Alternate'),
    (PROV_REC_SPECIALIZATION,       u'Specialization'),
    (PROV_REC_MENTION,              u'Mention'),
#    (PROV_REC_COLLECTION,           u'Collection'),
    (PROV_REC_MEMBERSHIP,           u'Membership'),
    (PROV_REC_BUNDLE,               u'Bundle'),
)

PROV_N_MAP = {
    PROV_REC_ENTITY:               u'entity',
    PROV_REC_ACTIVITY:             u'activity',
    PROV_REC_GENERATION:           u'wasGeneratedBy',
    PROV_REC_USAGE:                u'used',
    PROV_REC_COMMUNICATION:        u'wasInformedBy',
    PROV_REC_START:                u'wasStartedBy',
    PROV_REC_END:                  u'wasEndedBy',
    PROV_REC_INVALIDATION:         u'wasInvalidatedBy',
    PROV_REC_DERIVATION:           u'wasDerivedFrom',
    PROV_REC_AGENT:                u'agent',
    PROV_REC_ATTRIBUTION:          u'wasAttributedTo',
    PROV_REC_ASSOCIATION:          u'wasAssociatedWith',
    PROV_REC_DELEGATION:           u'actedOnBehalfOf',
    PROV_REC_INFLUENCE:            u'wasInfluencedBy',
    PROV_REC_ALTERNATE:            u'alternateOf',
    PROV_REC_SPECIALIZATION:       u'specializationOf',
    PROV_REC_MENTION:              u'mentionOf',
    PROV_REC_COLLECTION:           u'collection',
    PROV_REC_MEMBERSHIP:           u'memberOf',
    PROV_REC_BUNDLE:               u'bundle',
}

## Identifiers for PROV's attributes
PROV_ATTR_ENTITY                = 1
PROV_ATTR_ACTIVITY              = 2
PROV_ATTR_TRIGGER               = 3
PROV_ATTR_INFORMED              = 4
PROV_ATTR_INFORMANT             = 5
PROV_ATTR_STARTER               = 6
PROV_ATTR_ENDER                 = 7
PROV_ATTR_AGENT                 = 8
PROV_ATTR_PLAN                  = 9
PROV_ATTR_DELEGATE              = 10
PROV_ATTR_RESPONSIBLE           = 11
PROV_ATTR_GENERATED_ENTITY      = 12
PROV_ATTR_USED_ENTITY           = 13
PROV_ATTR_GENERATION            = 14
PROV_ATTR_USAGE                 = 15
PROV_ATTR_SPECIFIC_ENTITY       = 16
PROV_ATTR_GENERAL_ENTITY        = 17
PROV_ATTR_ALTERNATE1            = 18
PROV_ATTR_ALTERNATE2            = 19
PROV_ATTR_BUNDLE                = 20
PROV_ATTR_INFLUENCEE            = 21
PROV_ATTR_INFLUENCER            = 22
PROV_ATTR_COLLECTION            = 23

# Literal properties
PROV_ATTR_TIME                  = 100
PROV_ATTR_STARTTIME             = 101
PROV_ATTR_ENDTIME               = 102

PROV_RECORD_ATTRIBUTES = (
    # Relations properties
    (PROV_ATTR_ENTITY, u'prov:entity'),
    (PROV_ATTR_ACTIVITY, u'prov:activity'),
    (PROV_ATTR_TRIGGER, u'prov:trigger'),
    (PROV_ATTR_INFORMED, u'prov:informed'),
    (PROV_ATTR_INFORMANT, u'prov:informant'),
    (PROV_ATTR_STARTER, u'prov:starter'),
    (PROV_ATTR_ENDER, u'prov:ender'),
    (PROV_ATTR_AGENT, u'prov:agent'),
    (PROV_ATTR_PLAN, u'prov:plan'),
    (PROV_ATTR_DELEGATE, u'prov:delegate'),
    (PROV_ATTR_RESPONSIBLE, u'prov:responsible'),
    (PROV_ATTR_GENERATED_ENTITY, u'prov:generatedEntity'),
    (PROV_ATTR_USED_ENTITY, u'prov:usedEntity'),
    (PROV_ATTR_GENERATION, u'prov:generation'),
    (PROV_ATTR_USAGE, u'prov:usage'),
    (PROV_ATTR_SPECIFIC_ENTITY, u'prov:specificEntity'),
    (PROV_ATTR_GENERAL_ENTITY, u'prov:generalEntity'),
    (PROV_ATTR_ALTERNATE1, u'prov:alternate1'),
    (PROV_ATTR_ALTERNATE2, u'prov:alternate2'),
    (PROV_ATTR_BUNDLE, u'prov:bundle'),
    (PROV_ATTR_INFLUENCEE, u'prov:influencee'),
    (PROV_ATTR_INFLUENCER, u'prov:influencer'),
    (PROV_ATTR_COLLECTION, u'prov:collection'),
    # Literal properties
    (PROV_ATTR_TIME, u'prov:time'),
    (PROV_ATTR_STARTTIME, u'prov:startTime'),
    (PROV_ATTR_ENDTIME, u'prov:endTime'),
)

PROV_ATTRIBUTE_LITERALS = set([PROV_ATTR_TIME, PROV_ATTR_STARTTIME, PROV_ATTR_ENDTIME])

PROV_RECORD_IDS_MAP = dict((PROV_N_MAP[rec_type_id], rec_type_id) for rec_type_id in PROV_N_MAP)
PROV_ID_ATTRIBUTES_MAP = dict((prov_id, attribute) for (prov_id, attribute) in PROV_RECORD_ATTRIBUTES)
PROV_ATTRIBUTES_ID_MAP = dict((attribute, prov_id) for (prov_id, attribute) in PROV_RECORD_ATTRIBUTES)

_r_xsd_dateTime = re.compile(""" ^
    (?P<year>-?[0-9]{4}) - (?P<month>[0-9]{2}) - (?P<day>[0-9]{2})
    T (?P<hour>[0-9]{2}) : (?P<minute>[0-9]{2}) : (?P<second>[0-9]{2})
    (?P<microsecond>\.[0-9]{1,6})?
    (?P<tz>
      Z | (?P<tz_hr>[-+][0-9]{2}) : (?P<tz_min>[0-9]{2})
    )?
    $ """, re.X)
_r_typed_literal_uri = re.compile(r'^"(?P<value>[^"\\]*(?:\\.[^"\\]*)*)"\^\^<(?P<datatype>[^>\\]*(?:\\.[^>\\]*)*)>$', re.X)
_r_typed_literal_qname = re.compile(r'^"(?P<value>[^"\\]*(?:\\.[^"\\]*)*)"\^\^(?P<datatype>[^>\\]*(?:\\.[^>\\]*)*)$', re.X)

# Datatypes
def _parse_xsd_dateTime(s):
    """Returns datetime or None."""
    m = _r_xsd_dateTime.match(s)
    if m is not None:
        values = m.groupdict()
    if values["microsecond"] is None:
        values["microsecond"] = 0
    else:
        values["microsecond"] = values["microsecond"][1:]
        values["microsecond"] += "0" * (6 - len(values["microsecond"]))
    values = dict((k, int(v)) for k, v in values.iteritems()
                  if not k.startswith("tz"))
    
    return datetime.datetime(**values)

def _ensure_datetime(time):
    if isinstance(time, basestring):
        return _parse_xsd_dateTime(time);
    else:
        return time

def parse_xsd_dateTime(s):
    try:
        return _parse_xsd_dateTime(s)
    except ValueError:
        pass
    return None
    
DATATYPE_PARSERS = {
    datetime.datetime: parse_xsd_dateTime,
}

def parse_datatype(value, datatype):
    if datatype in DATATYPE_PARSERS:
        # found the required parser
        return DATATYPE_PARSERS[datatype](value)
    else:
        # No parser found for the given data type
        raise Exception(u'No parser found for the data type <%s>' % str(datatype))
        
class Literal(object):
    def __init__(self, value, datatype=None, langtag=None):
        self._value = value
        self._datatype = datatype
        self._langtag = langtag
        
    def __str__(self):
        return self.provn_representation()
    
    def get_value(self):
        return self._value
    
    def get_datatype(self):
        return self._datatype
    
    def provn_representation(self):
        if self._langtag:
            # a langtag can only goes with string
            return u'"%s"@%s' % (str(self._value), str(self._langtag))
        else:
            return u'"%s" %%%% %s' % (str(self._value), str(self._datatype))
        
    def json_representation(self):
        if self._langtag:
            # a langtag can only goes with string
            return { '$': str(self._value), 'lang': self._langtag}
        else:
            if isinstance(self._datatype, QName):
                return { '$': str(self._value), 'type': str(self._datatype)}
            else:
                # Assuming it is a valid identifier
                return { '$': str(self._value), 'type': self._datatype.get_uri()}

class Identifier(object):
    def __init__(self, uri):
        self._uri = uri
        
    def get_uri(self):
        return self._uri
    
    def __str__(self):
        return self._uri
    
    def __eq__(self, other):
        return self.get_uri() == other.get_uri() if isinstance(other, Identifier) else False
    
    def __hash__(self):
        return hash(self.get_uri())
    
    def provn_representation(self):
        return u'"%s" %%%% xsd:anyURI' % self._uri
    
    def json_representation(self):
        return { '$': self._uri, 'type': u'xsd:anyURI'}
    

class QName(Identifier):
    def __init__(self, namespace, localpart):
        self._namespace = namespace
        self._localpart = localpart
        self._str = ':'.join([namespace._prefix, localpart]) if namespace._prefix else localpart 
        
    def get_namespace(self):
        return self._namespace
    
    def get_localpart(self):
        return self._localpart
    
    def get_uri(self):
        return ''.join([self._namespace._uri, self._localpart])
    
    def __str__(self):
        return self._str
    
    def provn_representation(self):
        return u"'%s'" % self._str
    
    def json_representation(self):
        return { '$': self._str, 'type': u'xsd:QName'}
    

class Namespace(object):
    def __init__(self, prefix, uri):
        self._prefix = prefix
        self._uri = uri
        self._cache = dict()

    def get_prefix(self):
        return self._prefix
            
    def get_uri(self):
        return self._uri
    
    def contains(self, identifier):
        uri = identifier if isinstance(identifier, (str, unicode)) else (identifier.get_uri() if isinstance(identifier, Identifier) else None)
        return uri.startswith(self._uri) if uri else False
    
    def qname(self, identifier):
        uri = identifier if isinstance(identifier, (str, unicode)) else (identifier.get_uri() if isinstance(identifier, Identifier) else None)
        if uri and uri.startswith(self._uri):
            return QName(self, uri[len(self._uri):])
        else:
            return None
        
    def __eq__(self, other):
        return (self._uri == other._uri and self._prefix == other._prefix) if isinstance(other, Namespace) else False
    
    def __getitem__(self, localpart):
        if localpart in self._cache:
            return self._cache[localpart]
        else:
            qname = QName(self, localpart)
            self._cache[localpart] = qname
            return qname
    
XSD = Namespace("xsd",'http://www.w3.org/2001/XMLSchema-datatypes#')
PROV = Namespace("prov",'http://www.w3.org/ns/prov#')
    
### Exceptions

class ProvException(Exception):
    """Base class for exceptions in this module."""
    pass

class ProvExceptionMissingRequiredAttribute(ProvException):
    def __init__(self, record_type, attribute_id):
        self.record_type = record_type
        self.attribute_id = attribute_id

class ProvExceptionNotValidAttribute(ProvException):
    def __init__(self, record_type, attribute, attribute_types):
        self.record_type = record_type
        self.attribute = attribute
        self.attribute_types = attribute_types
        
class ProvExceptionContraint(ProvException):
    def __init__(self, record_type, attribute1, attribute2, msg):
        self.record_type = record_type
        self.attribute1 = attribute1
        self.attribute2 = attribute2
        self.msg = msg

        
# PROV records
class ProvRecord(object):
    """Base class for PROV _records."""
    def __init__(self, bundle, identifier, attributes=None, other_attributes=None):
        self._bundle = bundle
        self._identifier = identifier
        self._attributes = None
        self._extra_attributes = None
        if attributes or other_attributes:
            self.add_attributes(attributes, other_attributes)
        
    def get_type(self):
        pass
    
    def get_prov_type(self):
        pass
    
    def get_asserted_types(self):
        if self._extra_attributes:
            prov_type = PROV['type']
            return set([value for attr, value in self._extra_attributes if attr == prov_type])
        return set()
    
    def add_asserted_type(self, type_identifier):
        asserted_types = self.get_asserted_types()
        if type_identifier not in asserted_types:
            if self._extra_attributes is None:
                self._extra_attributes = []
            self._extra_attributes.append((PROV['type'], type_identifier))
    
    def get_identifier(self):
        return self._identifier

    def get_label(self):
        label = None
        for attribute in self._extra_attributes:
            if attribute[0]:
                if attribute[0] == PROV['label']:
                    label = attribute[1]
                    # use the first label found
                    break
        return label if label else self._identifier 

    def add_extra_attributes(self, extra_attributes):
        if extra_attributes:
            if self._extra_attributes is None:
                self._extra_attributes = []
            try:
                # This will only work if extra_attributes is a dictionary
                # Converting the dictionary into a list of tuples (i.e. attribute-value pairs) 
                extra_attributes = extra_attributes.items()
            except:
                # Do nothing if it did not work, expect the variable is already a list
                pass
            attr_list = ((self._bundle.valid_identifier(attribute), value) for attribute, value in extra_attributes)
            # Check attributes for valid qualified names
            self._extra_attributes.extend(attr_list)
            
    def add_attributes(self, attributes, extra_attributes):
        if attributes:
            if self._attributes is None:
                self._attributes = attributes
            else:
                self._attributes.update(attributes)
        self.add_extra_attributes(extra_attributes)
    
    def get_attributes(self):
        return (self._attributes, self._extra_attributes)
    
    def get_bundle(self):
        return self._bundle
    
    def _parse_identifier(self, value):
        try:
            return value.get_identifier()
        except:
            return self._bundle.valid_identifier(value)
    
    def _parse_record(self, attribute, attribute_types):
        # check to see if there is an existing record matching the attribute (as the record's identifier)
        existing_record = self._bundle.get_record(attribute)
        if existing_record is None:
            existing_record = self._bundle.get_bundle(attribute)
        if existing_record and isinstance(existing_record, attribute_types):
            return existing_record
        else:
            return None
    
    def _parse_attribute(self, attribute, attribute_types):
        if attribute_types is Identifier:
            return self._parse_identifier(attribute)
        # attempt to find an existing record
        record = self._parse_record(attribute, attribute_types)
        if record:
            return record
        else:
            # It is not a record, try to parse it with known datatype parsers
            if isinstance(attribute_types, collections.Iterable):
                for datatype in attribute_types:
                    data = parse_datatype(attribute, datatype)
                    if data:
                        return data
            else:
                # only one datatype provided
                datatype = attribute_types
                try:
                    return parse_datatype(attribute, datatype)
                except:
                    return None
        return None        
    
    def _validate_attribute(self, attribute, attribute_types):
        if isinstance(attribute, attribute_types):
            # The attribute is of a required type
            # Return it
            return attribute
        else:
            # The attribute is not of a valid type
            # Attempt to parse it
            parsed_value = self._parse_attribute(attribute, attribute_types)
            if parsed_value is None:
                raise ProvExceptionNotValidAttribute(self.get_type(), attribute, attribute_types)
            return parsed_value
        
                
    def required_attribute(self, attributes, attribute_id, attribute_types):
        if attribute_id not in attributes:
            # Raise an exception about the missing attribute
            raise ProvExceptionMissingRequiredAttribute(self.get_type(), attribute_id)
        # Found the required attribute
        attribute = attributes.get(attribute_id)
        return self._validate_attribute(attribute, attribute_types)
            
    def optional_attribute(self, attributes, attribute_id, attribute_types):
        if attribute_id not in attributes:
            # Because this is optional, return nothing
            return None
        # Found the optional attribute
        attribute = attributes.get(attribute_id)
        if attribute is None:
            return None
        # Validate its type
        return self._validate_attribute(attribute, attribute_types)
    
                
    def __eq__(self, other):
        if self.__class__ <> other.__class__:
            return False
        if self._identifier and not (self._identifier == other._identifier):
            return False
        if self._attributes and other._attributes:
            if len(self._attributes) <> len(other._attributes):
                return False
            for attr, value_a in self._attributes.items():
                value_b = other._attributes[attr]
                if isinstance(value_a, ProvRecord) and value_a._identifier:
                    if not (value_a._identifier == value_b._identifier):
                        return False
                elif not (value_a == value_b):
                    return False
        elif other._attributes and not self._attributes:
            other_attrs = [(key, value) for key, value in other._attributes.items() if value is not None]
            if other_attrs:
                # the other's attributes set is not empty.
                return False
        elif self._attributes and not other._attributes:
            my_attrs = [(key, value) for key, value in self._attributes.items() if value is not None]
            if my_attrs:
                # my attributes set is not empty.
                return False
        if self._extra_attributes <> other._extra_attributes:
            return False
        return True 
          
    def __str__(self):
        return self.get_provn()
    
    def get_provn(self, _indent_level=0):
        items = []
        if self._identifier:
            items.append(str(self._identifier))
        if self._attributes:
            for (attr, value) in self._attributes.items():
                if value is None:
                    items.append('-')
                else:
                    if isinstance(value, ProvRecord):
                        record_id = value.get_identifier()
                        items.append(str(record_id))
                    else:
                        # Assuming this is a datetime or QName value
                        items.append(value.isoformat() if isinstance(value, datetime.datetime) else str(value))
        
        if self._extra_attributes:
            extra = []
            for (attr, value) in self._extra_attributes:
                try:
                    # try if there is a prov-n representation defined
                    provn_represenation = value.provn_representation()
                except:
                    if isinstance(value, basestring):
                        provn_represenation = '"%s"' % value
                    else:
                        # asssuming it is datetime, otherwise, fallback to str
                        provn_represenation = '"%s" %%%% xsd:dateTime' % value.isoformat() if isinstance(value, datetime.datetime) else str(value)
                extra.append('%s=%s' % (str(attr), provn_represenation))
            if extra:
                items.append('[%s]' % ', '.join(extra))
        
        return '%s(%s)' % (PROV_N_MAP[self.get_type()], ', '.join(items))
    
    def is_element(self):
        return False
    
    def is_relation(self):
        return False

class ProvElement(ProvRecord):
    def is_element(self):
        return True

class ProvRelation(ProvRecord):
    def is_relation(self):
        return True


### Component 1: Entities and Activities

class ProvEntity(ProvElement):
    def get_type(self):
        return PROV_REC_ENTITY
    
    def get_prov_type(self):
        return PROV['Entity']
    

class ProvActivity(ProvElement):
    def get_type(self):
        return PROV_REC_ACTIVITY
    
    def get_prov_type(self):
        return PROV['Activity']
    
    def add_attributes(self, attributes, extra_attributes):
        startTime = self.optional_attribute(attributes, PROV_ATTR_STARTTIME, datetime.datetime)
        endTime = self.optional_attribute(attributes, PROV_ATTR_ENDTIME, datetime.datetime)
        if startTime and endTime and startTime > endTime:
            #TODO Raise logic exception here
            pass
        attributes = OrderedDict()
        attributes[PROV_ATTR_STARTTIME] = startTime
        attributes[PROV_ATTR_ENDTIME] = endTime
            
        ProvElement.add_attributes(self, attributes, extra_attributes)

    # Convenient methods
    def set_time(self, startTime=None, endTime=None):
        # The _attributes dict should be initialised
        self._attributes[PROV_ATTR_STARTTIME] = startTime
        self._attributes[PROV_ATTR_ENDTIME] = endTime

class ProvGeneration(ProvRelation):
    def get_type(self):
        return PROV_REC_GENERATION
    
    def get_prov_type(self):
        return PROV['Generation']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        entity = self.required_attribute(attributes, PROV_ATTR_ENTITY, ProvEntity) 
        # Optional attributes
        activity = self.optional_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        time = self.optional_attribute(attributes, PROV_ATTR_TIME, datetime.datetime)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ENTITY] = entity 
        attributes[PROV_ATTR_ACTIVITY] = activity
        attributes[PROV_ATTR_TIME] = time
        
        ProvRelation.add_attributes(self, attributes, extra_attributes)


class ProvUsage(ProvRelation):
    def get_type(self):
        return PROV_REC_USAGE
    
    def get_prov_type(self):
        return PROV['Usage']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        activity = self.required_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        # Optional attributes
        entity = self.optional_attribute(attributes, PROV_ATTR_ENTITY, ProvEntity) 
        time = self.optional_attribute(attributes, PROV_ATTR_TIME, datetime.datetime)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ACTIVITY] = activity
        attributes[PROV_ATTR_ENTITY] = entity
        attributes[PROV_ATTR_TIME] = time
        ProvRelation.add_attributes(self, attributes, extra_attributes)

class ProvCommunication(ProvRelation):
    def get_type(self):
        return PROV_REC_COMMUNICATION
    
    def get_prov_type(self):
        return PROV['Communication']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        informed = self.required_attribute(attributes, PROV_ATTR_INFORMED, ProvActivity)
        informant = self.required_attribute(attributes, PROV_ATTR_INFORMANT, ProvActivity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_INFORMED] = informed
        attributes[PROV_ATTR_INFORMANT] = informant
        ProvRelation.add_attributes(self, attributes, extra_attributes)

class ProvStart(ProvRelation):
    def get_type(self):
        return PROV_REC_START
    
    def get_prov_type(self):
        return PROV['Start']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        activity = self.required_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        # Optional attributes
        trigger = self.optional_attribute(attributes, PROV_ATTR_TRIGGER, ProvEntity)
        starter = self.optional_attribute(attributes, PROV_ATTR_STARTER, ProvActivity)
        time = self.optional_attribute(attributes, PROV_ATTR_TIME, datetime.datetime)

        attributes = OrderedDict()
        attributes[PROV_ATTR_ACTIVITY] = activity
        attributes[PROV_ATTR_TRIGGER] = trigger
        attributes[PROV_ATTR_STARTER] = starter
        attributes[PROV_ATTR_TIME] = time
        ProvRelation.add_attributes(self, attributes, extra_attributes)
        
class ProvEnd(ProvRelation):
    def get_type(self):
        return PROV_REC_END
    
    def get_prov_type(self):
        return PROV['End']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        activity = self.required_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        # Optional attributes
        trigger = self.optional_attribute(attributes, PROV_ATTR_TRIGGER, ProvEntity)
        ender = self.optional_attribute(attributes, PROV_ATTR_ENDER, ProvActivity)
        time = self.optional_attribute(attributes, PROV_ATTR_TIME, datetime.datetime)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ACTIVITY] = activity
        attributes[PROV_ATTR_TRIGGER] = trigger
        attributes[PROV_ATTR_ENDER] = ender
        attributes[PROV_ATTR_TIME] = time
        ProvRelation.add_attributes(self, attributes, extra_attributes)


class ProvInvalidation(ProvRelation):
    def get_type(self):
        return PROV_REC_INVALIDATION
    
    def get_prov_type(self):
        return PROV['Invalidation']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        entity = self.required_attribute(attributes, PROV_ATTR_ENTITY, ProvEntity)
        # Optional attributes
        activity = self.optional_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        time = self.optional_attribute(attributes, PROV_ATTR_TIME, datetime.datetime)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ENTITY] = entity
        attributes[PROV_ATTR_ACTIVITY] = activity
        attributes[PROV_ATTR_TIME] = time
        ProvRelation.add_attributes(self, attributes, extra_attributes)


### Component 2: Derivations

class ProvDerivation(ProvRelation):
    def get_type(self):
        return PROV_REC_DERIVATION
    
    def get_prov_type(self):
        return PROV['Derivation']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        generatedEntity = self.required_attribute(attributes, PROV_ATTR_GENERATED_ENTITY, ProvEntity)
        usedEntity = self.required_attribute(attributes, PROV_ATTR_USED_ENTITY, ProvEntity)
        # Optional attributes
        activity = self.optional_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity) 
        generation = self.optional_attribute(attributes, PROV_ATTR_GENERATION, ProvGeneration)
        usage = self.optional_attribute(attributes, PROV_ATTR_USAGE, ProvUsage)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_GENERATED_ENTITY]= generatedEntity
        attributes[PROV_ATTR_USED_ENTITY]= usedEntity
        attributes[PROV_ATTR_ACTIVITY]= activity
        attributes[PROV_ATTR_GENERATION] = generation
        attributes[PROV_ATTR_USAGE] = usage
        ProvRelation.add_attributes(self, attributes, extra_attributes)


### Component 3: Agents, Responsibility, and Influence

class ProvAgent(ProvElement):
    def get_type(self):
        return PROV_REC_AGENT
    
    def get_prov_type(self):
        return PROV['Agent']

    
class ProvAttribution(ProvRelation):
    def get_type(self):
        return PROV_REC_ATTRIBUTION
    
    def get_prov_type(self):
        return PROV['Attribution']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        entity = self.required_attribute(attributes, PROV_ATTR_ENTITY, ProvEntity)
        agent = self.required_attribute(attributes, PROV_ATTR_AGENT, (ProvAgent, ProvEntity))
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ENTITY] = entity
        attributes[PROV_ATTR_AGENT] = agent
        ProvRelation.add_attributes(self, attributes, extra_attributes)

class ProvAssociation(ProvRelation):
    def get_type(self):
        return PROV_REC_ASSOCIATION
    
    def get_prov_type(self):
        return PROV['Association']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        activity = self.required_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity) 
        # Optional attributes
        agent = self.optional_attribute(attributes, PROV_ATTR_AGENT, (ProvAgent, ProvEntity))
        plan = self.optional_attribute(attributes, PROV_ATTR_PLAN, ProvEntity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ACTIVITY]= activity
        attributes[PROV_ATTR_AGENT]= agent
        attributes[PROV_ATTR_PLAN]= plan
        ProvRelation.add_attributes(self, attributes, extra_attributes)


class ProvDelegation(ProvRelation):
    def get_type(self):
        return PROV_REC_DELEGATION
    
    def get_prov_type(self):
        return PROV['Delegation']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        delegate = self.required_attribute(attributes, PROV_ATTR_DELEGATE, (ProvAgent, ProvEntity)) 
        responsible = self.required_attribute(attributes, PROV_ATTR_RESPONSIBLE, (ProvAgent, ProvEntity))
        # Optional attributes
        activity = self.optional_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_DELEGATE] = delegate
        attributes[PROV_ATTR_RESPONSIBLE] = responsible
        attributes[PROV_ATTR_ACTIVITY]= activity
        ProvRelation.add_attributes(self, attributes, extra_attributes)

class ProvInfluence(ProvRelation):
    def get_type(self):
        return PROV_REC_INFLUENCE
    
    def get_prov_type(self):
        return PROV['Influence']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        influencee = self.required_attribute(attributes, PROV_ATTR_INFLUENCEE, (ProvAgent, ProvEntity)) 
        influencer = self.required_attribute(attributes, PROV_ATTR_INFLUENCER, (ProvAgent, ProvEntity))
        # Optional attributes
        activity = self.optional_attribute(attributes, PROV_ATTR_ACTIVITY, ProvActivity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_INFLUENCEE] = influencee
        attributes[PROV_ATTR_INFLUENCER] = influencer
        attributes[PROV_ATTR_ACTIVITY]= activity
        ProvRelation.add_attributes(self, attributes, extra_attributes)


### Component 4: Bundles

# See below

### Component 5: Alternate Entities

class ProvSpecialization(ProvRelation):
    def get_type(self):
        return PROV_REC_SPECIALIZATION
    
    def get_prov_type(self):
        return PROV['Specialization']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        specificEntity = self.required_attribute(attributes, PROV_ATTR_SPECIFIC_ENTITY, ProvEntity) 
        generalEntity = self.required_attribute(attributes, PROV_ATTR_GENERAL_ENTITY, ProvEntity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_SPECIFIC_ENTITY]= specificEntity
        attributes[PROV_ATTR_GENERAL_ENTITY]= generalEntity
        ProvRelation.add_attributes(self, attributes, extra_attributes)


class ProvAlternate(ProvRelation):
    def get_type(self):
        return PROV_REC_ALTERNATE
    
    def get_prov_type(self):
        return PROV['Alternate']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        alternate1 = self.required_attribute(attributes, PROV_ATTR_ALTERNATE1, ProvEntity) 
        alternate2 = self.required_attribute(attributes, PROV_ATTR_ALTERNATE2, ProvEntity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_ALTERNATE1]= alternate1
        attributes[PROV_ATTR_ALTERNATE2]= alternate2
        ProvRelation.add_attributes(self, attributes, extra_attributes)

class ProvMention(ProvSpecialization):
    def get_type(self):
        return PROV_REC_MENTION
    
    def get_prov_type(self):
        return PROV['Mention']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        specificEntity = self.required_attribute(attributes, PROV_ATTR_SPECIFIC_ENTITY, ProvEntity) 
        generalEntity = self.required_attribute(attributes, PROV_ATTR_GENERAL_ENTITY, Identifier)
        bundle = self.required_attribute(attributes, PROV_ATTR_BUNDLE, Identifier)
        #=======================================================================
        # # This is disabled so that mentionOf can refer to bundle that is not defined in the same place
        # bundle = self.required_attribute(attributes, PROV_ATTR_BUNDLE, ProvBundle)
        # # Check if generalEntity is in the bundle
        # if generalEntity.get_bundle() is not bundle:
        #    raise ProvExceptionContraint(PROV_REC_MENTION, generalEntity, bundle, 'The generalEntity must belong to the bundle')
        #=======================================================================
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_SPECIFIC_ENTITY]= specificEntity
        attributes[PROV_ATTR_GENERAL_ENTITY]= generalEntity
        attributes[PROV_ATTR_BUNDLE]= bundle
        ProvRelation.add_attributes(self, attributes, extra_attributes)


### Component 6: Collections

class ProvCollection(ProvEntity):
    def get_type(self):
        return PROV_REC_COLLECTION
    
    def get_prov_type(self):
        return PROV['Collection']

class ProvMembership(ProvRelation):
    def get_type(self):
        return PROV_REC_MEMBERSHIP
    
    def get_prov_type(self):
        return PROV['Membership']
    
    def add_attributes(self, attributes, extra_attributes):
        # Required attributes
        collection = self.required_attribute(attributes, PROV_ATTR_COLLECTION, ProvEntity) 
        entity = self.required_attribute(attributes, PROV_ATTR_ENTITY, ProvEntity)
        
        attributes = OrderedDict()
        attributes[PROV_ATTR_COLLECTION]= collection
        attributes[PROV_ATTR_ENTITY]= entity
        ProvRelation.add_attributes(self, attributes, extra_attributes)

# Class mappings from PROV record type    
PROV_REC_CLS = {
    PROV_REC_ENTITY                 : ProvEntity,
    PROV_REC_ACTIVITY               : ProvActivity,
    PROV_REC_GENERATION             : ProvGeneration,
    PROV_REC_USAGE                  : ProvUsage,
    PROV_REC_COMMUNICATION          : ProvCommunication,
    PROV_REC_START                  : ProvStart,
    PROV_REC_END                    : ProvEnd,
    PROV_REC_INVALIDATION           : ProvInvalidation,
    PROV_REC_DERIVATION             : ProvDerivation,
    PROV_REC_AGENT                  : ProvAgent,
    PROV_REC_ATTRIBUTION            : ProvAttribution,
    PROV_REC_ASSOCIATION            : ProvAssociation,
    PROV_REC_DELEGATION             : ProvDelegation,
    PROV_REC_INFLUENCE              : ProvInfluence,
    PROV_REC_SPECIALIZATION         : ProvSpecialization,
    PROV_REC_ALTERNATE              : ProvAlternate,
    PROV_REC_MENTION                : ProvMention,
    PROV_REC_COLLECTION             : ProvCollection,
    PROV_REC_MEMBERSHIP             : ProvMembership,
}


# Bundle
class NamespaceManager(dict):
    def __init__(self, default_namespaces={}, default=None):
        self._default_namespaces = {}
        self._default_namespaces.update(default_namespaces)
        self._namespaces = {}
        self.update(self._default_namespaces)
        self._default = default
        # TODO check if default is in the default namespaces
        self._anon_id_count = 0 
        self._rename_map = {}
        
    def get_namespace(self, uri):
        for namespace in self.values():
            if uri == namespace._uri:
                return namespace
        return None
    
    def get_registered_namespaces(self):
        return self._namespaces.values()
    
    def set_default_namespace(self, uri):
        self._default = Namespace('', uri)
        self[''] = self._default  
        
    def get_default_namespace(self):
        return self._default
    
    def add_namespace(self, namespace):
        if namespace in self.values():
            # no need to do anything
            return
        if namespace in self._rename_map:
            # already renamed and added
            return
        
        prefix = namespace.get_prefix()
        if prefix in self:
            # Conflicting prefix
            new_prefix = self._get_unused_prefix(prefix)
            new_namespace = Namespace(new_prefix, namespace.get_uri())
            self._rename_map[namespace] = new_namespace
            prefix = new_prefix
            namespace = new_namespace
        self._namespaces[prefix] = namespace
        self[prefix] = namespace
    
    def get_valid_identifier(self, identifier):
        if not identifier:
            return None
        if isinstance(identifier, Identifier):
            if isinstance(identifier, QName):
                # Register the namespace if it has not been registered before
                namespace = identifier.get_namespace() 
                if namespace not in self.values():
                    self.add_namespace(namespace)
            # return the original identifier
            return identifier
        elif isinstance(identifier, (str, unicode)):
            if identifier.startswith('_:'):
                return None
            elif ':' in identifier:
                # check if the identifier contains a registered prefix
                prefix, local_part = identifier.split(':', 1)
                if prefix in self:
                    # return a new QName
                    return self[prefix][local_part]
                else:
                    # treat as a URI (with the first part as its scheme)
                    # check if the URI can be compacted
                    for namespace in self.values():
                        if identifier.startswith(namespace.get_uri()):
                            # create a QName with the namespace
                            return namespace[identifier.replace(namespace.get_uri(), '')] 
                    # return an Identifier with the given URI
                    return Identifier(identifier)
            elif self._default:
                # create and return an identifier in the default namespace
                return self._default[identifier]
            else:
                # TODO Should an exception raised here
                return Identifier(identifier) 
    
    def get_anonymous_identifier(self, local_prefix='id'):
        self._anon_id_count += 1
        return Identifier('_:%s%d' % (local_prefix, self._anon_id_count))
    
    def _get_unused_prefix(self, original_prefix):
        if original_prefix not in self:
            return original_prefix
        count = 1
        while True:
            new_prefix = '_'.join((original_prefix, count))
            if new_prefix in self:
                count += 1
            else:
                return new_prefix
        

class ProvBundle(ProvEntity):
    def __init__(self, bundle=None, identifier=None, attributes=None, other_attributes=None):
        # Initializing bundle-specific attributes
        self._records = list()
        self._id_map = dict()
        self._bundles = dict()
        if bundle is None:
            self._namespaces = NamespaceManager({ PROV.get_prefix(): PROV, XSD.get_prefix(): XSD})
        else:
            self._namespaces = bundle._namespaces
        
        # Initializing record-specific attributes
        super(ProvBundle, self).__init__(bundle, identifier, attributes, other_attributes)
        
    # Bundle configurations
    def set_default_namespace(self, uri):
        self._namespaces.set_default_namespace(uri)
        
    def get_default_namespace(self):
        return self._namespaces.get_default_namespace()
        
    def add_namespace(self, namespace_or_prefix, uri=None):
        if uri is None:
            self._namespaces.add_namespace(namespace_or_prefix)
        else:
            self._namespaces.add_namespace(Namespace(namespace_or_prefix, uri))

    def get_registered_namespaces(self):
        return self._namespaces.get_registered_namespaces()
        
    def valid_identifier(self, identifier):
        return self._namespaces.get_valid_identifier(identifier) 
        
    def get_anon_id(self, record):
        #TODO Implement a dict of self-generated anon ids for records without identifier
        return self._namespaces.get_anonymous_identifier()
    
    def get_records(self):
        return self._records
    
    def get_record(self, identifier):
        try:
            valid_id = self.valid_identifier(identifier)
            return self._id_map[valid_id]
        except:
            # looking up the parent bundle
            if self._bundle is not None:
                return self._bundle.get_record(valid_id)
            else:
                return None
    
    def get_bundle(self, identifier):
        try:
            valid_id = self.valid_identifier(identifier)
            return self._bundles[valid_id]
        except:
            # looking up the parent bundle
            if self._bundle is not None:
                return self._bundle.get_bundle(valid_id)
            else:
                return None
        
    # PROV-JSON serialization/deserialization    
    class JSONEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, ProvBundle):
                return o._encode_JSON_container()
            else:
                # Use the default encoder instead
                return json.JSONEncoder.default(self, o)
            
    class JSONDecoder(json.JSONDecoder):
        def decode(self, s):
            json_container = json.JSONDecoder.decode(self, s)
            result = ProvBundle()
            result._decode_JSON_container(json_container)
            return result
    
    def _encode_json_representation(self, value):
        try:
            return value.json_representation()
        except AttributeError:
            if isinstance(value, datetime.datetime):
                return { '$': value.isoformat(), 'type': u'xsd:dateTime'}
            else:
                return value
    
    def _decode_json_representation(self, literal):
        try:
            value = literal['$']
            if 'lang' in literal:
                return Literal(value, langtag=literal['lang'])
            else:
                datatype = literal['type']
                # TODO Add a proper XSD datatype converter to replace this, e.g. for integers or floats
                if datatype == u'xsd:anyURI':
                    return Identifier(value)
                elif datatype == u'xsd:QName':
                    return self.valid_identifier(value)
                elif datatype == u'xsd:dateTime':
                    return parse_xsd_dateTime(value)
                else:
                    return Literal(value, self.valid_identifier(datatype))
        except:
            # simple type, just return it
            return literal
        
    def _encode_JSON_container(self):
        container = defaultdict(dict)
        
        if self._bundle is None:
            # This is the top-level bundle, we need to define namespaces
            prefixes = {}
            for namespace in self._namespaces.get_registered_namespaces():
                prefixes[namespace.get_prefix()] = namespace.get_uri()
            if self._namespaces._default:
                prefixes['$'] = self._namespaces._default.get_uri()
            container[u'prefix'] = prefixes
            
        ids = {}
        # generating/mapping all record identifiers 
        for record in self._records:
            ids[record] = record._identifier if record._identifier else self.get_anon_id(record)
        for record in self._records:
            rec_type = record.get_type()
            rec_label = PROV_N_MAP[rec_type]
            identifier = str(ids[record])
            
            if rec_type == PROV_REC_BUNDLE:
                # encoding the sub-bundle
                record_json = record._encode_JSON_container()
            else:
                record_json = {}
                if record._attributes:
                    for (attr, value) in record._attributes.items():
                        if isinstance(value, ProvRecord):
                            attr_record_id = ids[value]
                            record_json[PROV_ID_ATTRIBUTES_MAP[attr]] = str(attr_record_id) 
                        elif value is not None:
                            # Assuming this is a datetime value
                            record_json[PROV_ID_ATTRIBUTES_MAP[attr]] = value.isoformat() if isinstance(value, datetime.datetime) else str(value)
                if record._extra_attributes:
                    for (attr, value) in record._extra_attributes:
                        attr_id = str(attr)
                        value_json = self._encode_json_representation(value)
                        if attr_id in record_json:
                            # Multi-value attribute
                            existing_value = record_json[attr_id]
                            try:
                                # Add the value to the current list of values
                                existing_value.add(value_json)
                            except:
                                # But if the existing value is not a list, it'll fail
                                # create the list for the existing value and the second value
                                record_json[attr_id] = [existing_value, value_json]
                        else:
                            record_json[attr_id] = value_json
            container[rec_label][identifier] = record_json
        
        return container
    
    def _decode_JSON_container(self, jc):
        if u'prefix' in jc:
            prefixes = jc[u'prefix']
            for prefix, uri in prefixes.items():
                if prefix <> '$':
                    self.add_namespace(Namespace(prefix, uri))
                else:
                    self.set_default_namespace(uri)
        records = sorted([(PROV_RECORD_IDS_MAP[rec_type], rec_id, jc[rec_type][rec_id])
                          for rec_type in jc if rec_type <> u'prefix'
                          for rec_id in jc[rec_type]],
                         key=lambda tuple_rec: tuple_rec[0])
        
        record_map = {}
        # Create all the records before setting their attributes
        for (record_type, identifier, content) in records:
            if record_type == PROV_REC_BUNDLE:
                bundle = self.bundle(identifier)
                bundle._decode_JSON_container(content)
            else:
                record_map[identifier] = self.add_record(record_type, identifier, None, None)
        for (record_type, identifier, attributes) in records:
            if record_type <> PROV_REC_BUNDLE:
                record = record_map[identifier]
                
                if hasattr(attributes, 'items'): # it is a dict
                    # There is only one element, create a singleton list
                    elements = [attributes] 
                else: # expect it to be a list
                    # There are more than one element
                    # TODO: Fix this, we only accept one element (the first one)
                    elements = [attributes[0]]
                    
                for element in elements:    
                    prov_attributes = {}
                    extra_attributes = []
                    # Splitting PROV attributes and the others
                    for attr, value in element.items():
                        if attr in PROV_ATTRIBUTES_ID_MAP:
                            prov_attributes[PROV_ATTRIBUTES_ID_MAP[attr]] = record_map[value] if (isinstance(value, (str, unicode)) and value in record_map) else self._decode_json_representation(value)
                        else:
                            attr_id = self.valid_identifier(attr)
                            if isinstance(value, list):
                                # Parsing multi-value attribute
                                extra_attributes.append((attr_id, self._decode_json_representation(value_single)) for value_single in value)
                            else:
                                # add the single-value attribute
                                extra_attributes.append((attr_id, self._decode_json_representation(value)))
                    # TODO: This won't work when there are more than one element
                    record.add_attributes(prov_attributes, extra_attributes)
        
    # Miscellaneous functions
    def get_type(self):
        return PROV_REC_BUNDLE
    
    def get_provn(self, _indent_level=0):
        indentation = '' +  ('  ' * _indent_level)
        newline = '\n' + ('  ' * (_indent_level + 1))
        if self._bundle is None:
            records = ['bundle']
            # print out prefixes in the top-level bundle
            records.extend(['prefix %s <%s>' % (namespace.get_prefix(), namespace.get_uri()) for namespace in self._namespaces.get_registered_namespaces()])
            # a blank line between the prefixes and the assertions
            records.append('')
        else:
            records = ['bundle %s' % self._identifier]
        
        records.extend([record.get_provn(_indent_level + 1) for record in self._records])
        provn_str = newline.join(records) + '\n' + \
                    indentation + 'endBundle'
        return provn_str
        
    def __eq__(self, other):
        try:
            other_records = set(other._records)
        except:
            # other is not a bundle
            return False
        this_records = set(self._records)
        if len(this_records) <> len(other_records):
            return False
        # check if all records for equality
        for record_a in this_records:
            if record_a._identifier:
                if record_a.get_type() == PROV_REC_BUNDLE:
                    record_b = other.get_bundle(record_a._identifier)
                else:
                    record_b = other.get_record(record_a._identifier)
                if record_b: 
                    if record_a == record_b:
                        other_records.remove(record_b)
                        continue
                    else:
                        # XXX: temp_disable logger.debug("Inequal PROV records:")
                        # XXX: temp_disable logger.debug("%s" % str(record_a))
                        # XXX: temp_disable logger.debug("%s" % str(record_b))
                        return False
                else:
                    # XXX: temp_disable logger.debug("Could not find a record with this identifier: %s" % str(record_a._identifier))
                    return False
            else:
                # Manually look for the record
                found = False
                for record_b in other_records:
                    if record_a == record_b:
                        other_records.remove(record_b)
                        found = True
                        break
                if not found:
                    # XXX: temp_disable logger.debug("Could not find this record: %s" % str(record_a))
                    return False
        return True
            
    # Provenance statements
    def add_record(self, record_type, identifier, attributes=None, other_attributes=None):
        new_record = PROV_REC_CLS[record_type](self, self.valid_identifier(identifier), attributes, other_attributes)
        self._records.append(new_record)
        if new_record._identifier:
            if record_type == PROV_REC_BUNDLE:
                # Don't mix bunle ids with normal record ids.
                self._bundles[new_record._identifier] = new_record
            else:
                self._id_map[new_record._identifier] = new_record
        return new_record
    
        
    def add_bundle(self, identifier, bundle):
        '''Add a sub-bundle to the current bundle
        '''
        valid_id = self.valid_identifier(identifier)
        self._bundles[valid_id] = bundle
        bundle._bundle = self
        # TODO: Check namespace duplications, existing identifier
    
    def add_element(self, record_type, identifier, attributes=None, other_attributes=None):
        return self.add_record(record_type, identifier, attributes, other_attributes)
        
    def entity(self, identifier, other_attributes=None):
        return self.add_element(PROV_REC_ENTITY, identifier, None, other_attributes)
    
    def activity(self, identifier, startTime=None, endTime=None, other_attributes=None):
        return self.add_element(PROV_REC_ACTIVITY, identifier, {PROV_ATTR_STARTTIME: _ensure_datetime(startTime), PROV_ATTR_ENDTIME: _ensure_datetime(endTime)}, other_attributes)
    
    def generation(self, entity, activity=None, time=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_GENERATION, identifier, {PROV_ATTR_ENTITY: entity, PROV_ATTR_ACTIVITY: activity, PROV_ATTR_TIME: _ensure_datetime(time)}, other_attributes)
    
    def usage(self, activity, entity=None, time=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_USAGE, identifier, {PROV_ATTR_ACTIVITY: activity, PROV_ATTR_ENTITY: entity, PROV_ATTR_TIME: _ensure_datetime(time)}, other_attributes)
    
    def start(self, activity, trigger=None, starter=None, time=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_START, identifier, {PROV_ATTR_ACTIVITY: activity, PROV_ATTR_TRIGGER: trigger, PROV_ATTR_STARTER: starter, PROV_ATTR_TIME: _ensure_datetime(time)}, other_attributes)
    
    def end(self, activity, trigger=None, ender=None, time=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_END, identifier, {PROV_ATTR_ACTIVITY: activity, PROV_ATTR_TRIGGER: trigger, PROV_ATTR_ENDER: ender, PROV_ATTR_TIME: _ensure_datetime(time)}, other_attributes)
    
    def invalidation(self, entity, activity=None, time=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_INVALIDATION, identifier, {PROV_ATTR_ENTITY: entity, PROV_ATTR_ACTIVITY: activity, PROV_ATTR_TIME: _ensure_datetime(time)}, other_attributes)
    
    def communication(self, informed, informant, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_COMMUNICATION, identifier, {PROV_ATTR_INFORMED: informed, PROV_ATTR_INFORMANT: informant}, other_attributes)
    
    def agent(self, identifier, other_attributes=None):
        return self.add_element(PROV_REC_AGENT, identifier, None, other_attributes)
    
    def attribution(self, entity, agent, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_ATTRIBUTION, identifier, {PROV_ATTR_ENTITY: entity, PROV_ATTR_AGENT: agent}, other_attributes)
    
    def association(self, activity, agent=None, plan=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_ASSOCIATION, identifier, {PROV_ATTR_ACTIVITY: activity, PROV_ATTR_AGENT: agent, PROV_ATTR_PLAN: plan}, other_attributes)
    
    def delegation(self, delegate, responsible, activity=None, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_DELEGATION, identifier, {PROV_ATTR_DELEGATE: delegate, PROV_ATTR_RESPONSIBLE: responsible, PROV_ATTR_ACTIVITY: activity}, other_attributes)
        
    def influence(self, influencee, influencer, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_INFLUENCE, identifier, {PROV_ATTR_INFLUENCEE: influencee, PROV_ATTR_INFLUENCER: influencer}, other_attributes)
        
    def derivation(self, generatedEntity, usedEntity, activity=None, generation=None, usage=None, time=None, identifier=None, other_attributes=None):
        attributes = {PROV_ATTR_GENERATED_ENTITY: generatedEntity,
                      PROV_ATTR_USED_ENTITY: usedEntity,
                      PROV_ATTR_ACTIVITY: activity,
                      PROV_ATTR_GENERATION: generation,
                      PROV_ATTR_USAGE: usage}
        return self.add_record(PROV_REC_DERIVATION, identifier, attributes, other_attributes)
    
    def revision(self, generatedEntity, usedEntity, activity=None, generation=None, usage=None, time=None, identifier=None, other_attributes=None):
        record = self.derivation(generatedEntity, usedEntity, activity, generation, usage, time, identifier, other_attributes)
        record.add_asserted_type(PROV['Revision'])
        return record
    
    def quotation(self, generatedEntity, usedEntity, activity=None, generation=None, usage=None, time=None, identifier=None, other_attributes=None):
        record = self.derivation(generatedEntity, usedEntity, activity, generation, usage, time, identifier, other_attributes)
        record.add_asserted_type(PROV['Quotation'])
        return record
    
    def primary_source(self, generatedEntity, usedEntity, activity=None, generation=None, usage=None, time=None, identifier=None, other_attributes=None):
        record = self.derivation(generatedEntity, usedEntity, activity, generation, usage, time, identifier, other_attributes)
        record.add_asserted_type(PROV['PrimarySource'])
        return record
    
    def specialization(self, specificEntity, generalEntity, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_SPECIALIZATION, identifier, {PROV_ATTR_SPECIFIC_ENTITY: specificEntity, PROV_ATTR_GENERAL_ENTITY: generalEntity}, other_attributes)
    
    def alternate(self, alternate1, alternate2, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_ALTERNATE, identifier, {PROV_ATTR_ALTERNATE1: alternate1, PROV_ATTR_ALTERNATE2: alternate2}, other_attributes)
    
    def mention(self, specificEntity, generalEntity, bundle, identifier=None, other_attributes=None):
        # Special case for mention: if generalEntity is not already a ProvRecord (e.g. a string identifier),
        # then it might need to be looked up in the bundle provided, not the current bundle
        if isinstance(generalEntity, basestring):
            # try to look for an existing record first:
            existing_record = self.get_record(generalEntity)
            if existing_record:
                # stop looking
                generalEntity = existing_record
            else:
                # looking in the target bundle
                if isinstance(bundle, basestring):
                    bundle = self.get_bundle(bundle)
                generalEntity = bundle.get_record(generalEntity)
        return self.add_record(PROV_REC_MENTION, identifier, {PROV_ATTR_SPECIFIC_ENTITY: specificEntity, PROV_ATTR_GENERAL_ENTITY: generalEntity, PROV_ATTR_BUNDLE: bundle}, other_attributes)
    
    def collection(self, identifier, other_attributes=None):
        return self.add_element(PROV_REC_COLLECTION, identifier, None, other_attributes)
    
    def membership(self, collection, entity, identifier=None, other_attributes=None):
        return self.add_record(PROV_REC_MEMBERSHIP, identifier, {PROV_ATTR_COLLECTION: collection, PROV_ATTR_ENTITY: entity}, other_attributes)
    
    def bundle(self, identifier, other_attributes=None):
        return self.add_element(PROV_REC_BUNDLE, identifier, None, other_attributes)
    
    # Aliases
    wasGeneratedBy = generation
    used = usage
    wasStartedBy = start
    wasEndedBy = end
    wasInvalidatedBy = invalidation
    wasInformedBy = communication
    wasAttributedTo = attribution
    wasAssociatedWith = association
    actedOnBehalfOf = delegation
    wasInfluencedBy = influence
    wasDerivedFrom = derivation
    wasRevisionOf = revision
    wasQuotedFrom = quotation
    hadPrimarySource = primary_source
    alternateOf = alternate
    specializationOf = specialization
    mentionOf = mention
    hadMember = membership

# Add the newly defined ProvBundle into the PROV class dictionary    
PROV_REC_CLS[PROV_REC_BUNDLE] = ProvBundle



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Export data from several SPECs into a table.
"""

__docformat__ = 'restructuredtext'

# magic line for manpage summary
# man: -*- % generate a test SPEC from an arbitrary command call

import argparse
from ..spec import SPEC

parser_args = dict(formatter_class=argparse.RawDescriptionHelpFormatter)

def setup_parser(parser):
    parser.add_argument('specs', nargs='+', metavar='SPEC',
            help="SPEC filenames")
    parser.add_argument('-c', '--column', action='append', nargs='+',
            default=list(), required=True, metavar='ARG',
            help="column definition (data field, name, transform)")
    parser.add_argument('--csv', metavar='FILENAME',
            help='store table in CSV format into a file')
    parser.add_argument('--numtxt', metavar='FILENAME',
            help='store table in TXT format into a file')

def op2fx(op):
    if op == 'v':
        # unmodified
        return lambda x: x
    if op == 'len':
        return len
    else:
        raise ValueError("unknown operator")

def parse_column_def(coldef):
    ptr = header = None
    transform = lambda x: x
    deflen = len(coldef)
    ptr = coldef[0].split('->')
    # TODO implement a way to access list elements too
    if deflen > 1:
        header = coldef[1]
    if deflen > 2:
        transform = op2fx(coldef[2])
    return ptr, header, transform

def get_nested_value(dct, ptr, entities):
    if len(ptr):
        if ptr[0] == 'entity':
            # look whether an entity is references (implicitely)
            entity = dct['entity']
            if entity in entities:
                # reroute
                ptr[0] = entity
                dct = entities
        elif isinstance(dct, list) and ptr[0].isdigit():
            # index a list
            entity = dct[int(ptr[0])]
            if entity in entities:
                # reroute
                ptr[0] = entity
                dct = entities
        elif isinstance(dct, basestring):
            if dct in entities:
                # reroute
                ptr = [dct] + ptr
                dct = entities
        return get_nested_value(dct[ptr[0]], ptr[1:], entities)
    else:
        return dct

def export_as_csv(fname, rows, hdrs):
    import csv
    writer = csv.writer(open(fname, 'wb'), quoting=csv.QUOTE_NONNUMERIC)
    writer.writerow(hdrs)
    writer.writerows(rows)

def export_as_txt(fname, rows):
    from numpy import savetxt
    savetxt(fname, rows)

def run(args):
    coldefs = [parse_column_def(cd) for cd in args.column]
    # column headers
    hdrs = [c[1] for c in coldefs]
    rows = []
    for spec_fname in args.specs:
        spec = SPEC(open(spec_fname))
        row = [None] * len(coldefs)
        for i, c in enumerate(coldefs):
            try:
                val = get_nested_value(spec, c[0], spec.get('entities', {}))
                # transform
                val = c[2](val)
            except KeyError:
                print "Skipping '%s' in '%s' (field not found)" % ('->'.join(c[0]), spec_fname)
                val = None
            row[i] = val
        rows.append(row)
    if not args.csv is None:
        export_as_csv(args.csv, rows, hdrs)
    if not args.numtxt is None:
        try:
            export_as_txt(args.numtxt, rows)
        except TypeError:
            raise ValueError("failed to write an unsupported data type into '%s'" % args.numtxt)
    # TODO pytables/pandas support





'''Graphical visualisation support for prov.model.

This module produces graphical visualisation for provenanve graphs.
Requires pydot module and Graphviz.

References:

pydot Homepage: http://code.google.com/p/pydot/
Graphviz:       http://www.graphviz.org/
DOT Language:   http://www.graphviz.org/doc/info/lang.html

@author: Dong Huynh <trungdong@donggiang.com>
'''
from .prov import (ProvBundle, ProvElement,
                   PROV_REC_ACTIVITY, PROV_REC_AGENT,
                   PROV_REC_ALTERNATE, PROV_REC_ASSOCIATION,
                   PROV_REC_ATTRIBUTION, PROV_REC_BUNDLE,
                   PROV_REC_COMMUNICATION, PROV_REC_DERIVATION,
                   PROV_REC_DELEGATION, PROV_REC_ENTITY, PROV_REC_GENERATION,
                   PROV_REC_INFLUENCE, PROV_REC_INVALIDATION, PROV_REC_END,
                   PROV_REC_MEMBERSHIP, PROV_REC_MENTION,
                   PROV_REC_SPECIALIZATION, PROV_REC_START, PROV_REC_USAGE,
                   PROV_REC_COLLECTION)
#from ..utils.misc import package_check
#package_check('pydot')
import pydot

# Visual styles for various elements (nodes) and relations (edges)
# see http://graphviz.org/content/attrs
DOT_PROV_STYLE = {
    # Elements
    PROV_REC_ENTITY: {'shape': 'oval', 'style': 'filled', 'fillcolor': 'aliceblue'},
    PROV_REC_ACTIVITY: {'shape': 'box', 'style': 'filled', 'fillcolor': 'lemonchiffon'},
    PROV_REC_AGENT: {'shape': 'house', 'style': 'filled', 'fillcolor': 'yellow'},
    PROV_REC_COLLECTION: {'shape': 'oval'},
    PROV_REC_BUNDLE: {'shape': 'folder', 'style': 'filled', 'fillcolor': 'aliceblue'},
    # Relations
    PROV_REC_GENERATION: {'label': 'wasGeneratedBy', 'fontsize': '10.0', 'color': 'darkgreen', 'fontcolor': 'darkgreen'},
    PROV_REC_USAGE: {'label': 'used', 'fontsize': '10.0', 'color': 'red4', 'fontcolor': 'red'},
    PROV_REC_COMMUNICATION: {'label': 'wasInformedBy', 'fontsize': '10.0'},
    PROV_REC_START: {'label': 'wasStartedBy', 'fontsize': '10.0'},
    PROV_REC_END: {'label': 'wasEndedBy', 'fontsize': '10.0'},
    PROV_REC_INVALIDATION: {'label': 'wasInvalidatedBy', 'fontsize': '10.0'},
    PROV_REC_DERIVATION: {'label': 'wasDerivedFrom', 'fontsize': '10.0'},
    PROV_REC_ATTRIBUTION: {'label': 'wasAttributedTo', 'fontsize': '10.0', 'color': 'gold'},
    PROV_REC_ASSOCIATION: {'label': 'wasAssociatedWith', 'fontsize': '10.0', 'color': 'gold'},
    PROV_REC_DELEGATION: {'label': 'actedOnBehalfOf', 'fontsize': '10.0', 'color': 'gold'},
    PROV_REC_INFLUENCE: {'label': 'wasInfluencedBy', 'fontsize': '10.0', 'color': 'grey'},
    PROV_REC_ALTERNATE: {'label': 'alternateOf', 'fontsize': '10.0'},
    PROV_REC_SPECIALIZATION: {'label': 'specializationOf', 'fontsize': '10.0'},
    PROV_REC_MENTION: {'label': 'mentionOf', 'fontsize': '10.0'},
    PROV_REC_MEMBERSHIP: {'label': 'hadMember', 'fontsize': '10.0'},
    }

def prov_to_dot(prov_g, show_nary=False, use_labels=False):
    maindot = pydot.Dot(graph_type='digraph', rankdir='BT')

    node_map = {}
    count = [0, 0, 0]

    def _bundle_to_dot(dot, bundle):
        records = bundle.get_records()
        relations = []
        for rec in records:
            if rec.is_element():
                if isinstance(rec, ProvBundle):
                    count[2] = count[2] + 1
                    subdot = pydot.Cluster(graph_name='c%d' % count[2])
                    if use_labels:
                        subdot.set_label('"%s"' % str(rec.get_label()))
                    else:
                        subdot.set_label('"%s"' % str(rec.get_identifier()))
                    _bundle_to_dot(subdot, rec)
                    dot.add_subgraph(subdot)
                else:
                    count[0] = count[0] + 1
                    node_id = 'n%d' % count[0]
                    if use_labels:
                        node_label = '"%s"' % str(rec.get_label())
                    else:
                        node_label = '"%s"' % str(rec.get_identifier())
                    style = DOT_PROV_STYLE[rec.get_type()]
                    node = pydot.Node(node_id, label=node_label, **style)
                    node_map[rec] = node
                    dot.add_node(node)
            else:
                relations.append(rec)
        for rec in relations:
            nodes = [node for node in rec._attributes.values() if node is not None and isinstance(node, ProvElement)]
            if len(nodes) < 2:
                # Cannot draw this
                pass  
            elif len(nodes) == 2 or not show_nary:
                # binary relations
                style = DOT_PROV_STYLE[rec.get_type()]
                dot.add_edge(pydot.Edge(node_map[nodes[0]], node_map[nodes[1]], **style))
            else:
                # n-ary relations
                style = DOT_PROV_STYLE[rec.get_type()]
                # add a blank node
                count[1] = count[1] + 1
                bnode_id = 'b%d' % count[1]
                bnode = pydot.Node(bnode_id, label='""', shape='point')
                dot.add_node(bnode)

                dot.add_edge(pydot.Edge(node_map[nodes[0]], bnode, arrowhead='none', **style))
                style = dict(style)
                del style['label']
                for node in nodes[1:]:
                    dot.add_edge(pydot.Edge(bnode, node_map[node], **style))
                    style['color'] = 'gray'

    _bundle_to_dot(maindot, prov_g)
    return maindot

def prov_to_file(prov_g, filepath, use_labels=False, format='png',
                         dpi=150):
    """Write a prov json object to an image file
    """
    # Convert it to DOT
    dot = prov_to_dot(prov_g, use_labels=use_labels)
    dot.set_dpi(dpi)
    dot.write(filepath, format=format)
    return dot

### Testing code
if __name__=="__main__":
    import tempfile, os
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg
    import test.examples as ex

    path = tempfile.mkdtemp()
    filepath = os.path.join(path, 'dot-test.png')

    # Get an example PROV graph
    prov_g = ex.w3c_publication_1()
    # Convert it to DOT
    dot = prov_to_file(prov_g, filepath)

    # Display it using matplotlib
    img=mpimg.imread(filepath)
    imgplot = plt.imshow(img)
    plt.show()
    os.remove(filepath)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

from testkraut.config import ConfigManager
from nose.tools import *

def test_config():
    cfg = ConfigManager()
    # for now just running it a bit

    # query for some non-existing option and check if default is returned
    query = cfg.get('dasgibtsdochnicht', 'neegarnicht', default='required')
    assert_equal(query, 'required')




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

__docformat__ = 'restructuredtext'

import os
from os.path import join as opj
from testkraut import utils
from testkraut.pkg_mngr import PkgManager
from nose.tools import *
from testkraut.tests.test_runner import with_tempdir

def test_sysinfo():
    sysinfo = utils.describe_system()
    assert_true('python_version' in sysinfo)

def test_pkg_mngr():
    pkg = PkgManager()
    pypkgname = pkg.get_pkg_name('/usr/bin/python')
    if pkg.get_platform_name() in ('deb', 'rpm'):
        assert_false(pypkgname is None)
    pkg_info = pkg.get_pkg_info(pkg.get_pkg_name('/usr/bin/python'))

@with_tempdir()
def test_strace_wrapper(wdir):
    curdir = os.path.realpath(os.curdir)
    try:
        os.chdir(wdir)
        inf_path = opj(wdir, 'inf')
        outf_path = opj(wdir, 'outf')
        inf = open(inf_path, 'w')
        inf.write('TEST')
        inf.close()
        cmd = ['python', '-c', 'open(\"%s\",\"w\").write(open(\"%s\").read())'
                               % (outf_path, inf_path)]
        procs, retval = utils.get_cmd_prov_strace(cmd)
        assert_equal(retval, 0)
        assert_equal(len(procs), 1)
        assert_equal(procs.keys(), ['mother'])
        exe = procs['mother']
        assert_equal(exe['uses'], set(['inf']))
        assert_equal(exe['generates'], set(['outf']))
        assert_equal(exe['started_by'], None)
    finally:
        os.chdir(curdir)

#def test_debian_stuff():
#    try:
#        import platform
#        dist = platform.linux_distribution()[0]
#        if not dist in ['debian']:
#            raise nose.SkipTest
#    except:
#        raise nose.SkipTest
#    # I hope that most systems have this...
#    assert_equal(utils.get_debian_pkgname('/etc/deluser.conf'), 'adduser')





# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Cache input files requires by test(s).

"""

__docformat__ = 'restructuredtext'

# magic line for manpage summary
# man: -*- % generate a test SPEC from an arbitrary command call

import argparse
import os
import shutil
from os.path import join as opj
from glob import glob
from ..spec import SPEC
from ..utils import sha1sum, get_test_library_paths, get_spec

parser_args = dict(formatter_class=argparse.RawDescriptionHelpFormatter)

def setup_parser(parser):
    parser.add_argument('ids', nargs='*', metavar='ID',
            help="SPEC name/identifier")
    parser.add_argument('-l', '--library', action='append', default=[],
            help="""path to an additional test library""")
    parser.add_argument('-c', '--cache', default='filecache',
            help="path to the file cache")
    parser.add_argument('-s', '--search', action='append', default=[],
            help="where to search for files")
    parser.add_argument('--copy', action='store_true',
            help="copy files into the cache instead of symlinking them")

def run(args):
    lgr = args.logger
    if not len(args.ids):
        # if none specified go through all the SPECs in the lib
        args.ids = []
        for tld in get_test_library_paths(args.library): 
            args.ids.extend([os.path.basename(d)
                                for d in glob(opj(tld, '*'))
                                        if os.path.isdir(d)])
    wanted_files = set()
    hash_lookup = {}
    # scan the SPECs of all tests for needed files and their sha1sums
    for test_id in args.ids:
        lgr.debug("scan required files for test '%s'" % test_id)
        spec = get_spec(test_id, args.library)
        for _, input in spec.get_inputs('file').iteritems():
            if 'sha1sum' in input:
                wanted_files.add(input['sha1sum'])
                hash_lookup[input['sha1sum']] = (test_id, input.get('value', ''))
                lgr.debug("add '%s' (%s) to the list of files to look for"
                          % (input.get('value', ''), input['sha1sum']))
    # what do we have in the cache?
    have_files = [os.path.basename(f) for f in glob(opj(args.cache, '*')) if os.path.isfile(f)]
    # what is missing
    missing_files = wanted_files.difference(have_files)
    search_cache = {}
    # search in all dirs
    for search_dir in (args.search + args.library):
        for root, dirnames, filenames in os.walk(search_dir):
            for fname in filenames:
                fpath = opj(root, fname)
                sha1 = sha1sum(fpath)
                if sha1 in missing_files:
                    # make path relative to cache dir
                    search_cache[sha1] = os.path.relpath(fpath, args.cache)
                    lgr.debug("found missing '%s' at '%s'" % (sha1, fpath))
    # ensure the cache is there
    if not os.path.exists(args.cache):
        os.makedirs(args.cache)
    # copy/link them into the cache
    for sha1, fpath in search_cache.iteritems():
        dst_path = opj(args.cache, sha1)
        if os.path.lexists(dst_path):
            if os.path.islink(dst_path):
                # remove existing symlink
                os.remove(dst_path)
                lgr.debug("removing existing symlink '%s' in filecache"
                          % dst_path)
            else:
                lgr.warning(
                    "Will not replace existing non-symlink cache content: [%s: %s (%s)]"
                    % (hash_lookup[sha1] + (sha1,)))
        if not args.copy:
            os.symlink(fpath, dst_path)
            lgr.debug("symlink '%s'->'%s'" % (fpath, dst_path))
        elif hasattr(os, 'link'):
            # be nice and try hard-linking
            try:
                os.link(fpath, dst_path)
                lgr.debug("hardlink '%s'->'%s'" % (fpath, dst_path))
            except OSError:
                # silently fail if linking doesn't work (e.g.
                # cross-device link ... will recover later
                shutil.copy(fpath, dst_path)
                lgr.debug("copylink '%s'->'%s'" % (fpath, dst_path))
        else:
            shutil.copy(fpath, dst_path)
            lgr.debug("copylink '%s'->'%s'" % (fpath, dst_path))
        missing_files.remove(sha1)
    if len(missing_files):
        lgr.warning('cannot find needed file(s):')
        for mf in missing_files:
            lgr.warning('  %s: %s (%s)' % (hash_lookup[mf] + (mf,)))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the testkraut package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""""""

import os
import sys
from os.path import join as opj
import testkraut
from distutils.core import setup
from glob import glob


__docformat__ = 'restructuredtext'

extra_setuptools_args = {}
if 'setuptools' in sys.modules:
    extra_setuptools_args = dict(
        tests_require=['nose'],
        test_suite='nose.collector',
        zip_safe=False,
        extras_require = dict(
            doc='Sphinx>=0.3',
            test='nose>=0.10.1')
    )

test_library_files = []
for test in glob(opj('testkraut', 'library', '*')):
    if not os.path.isdir(test):
        continue
    for f in glob(opj(test, '*')):
        if not os.path.isfile(f):
            continue
        test_library_files.append(f[10:])

def main(**extra_args):
    setup(name         = 'testkraut',
          version      = testkraut.__version__,
          author       = 'Michael Hanke and the testkraut developers',
          author_email = 'michael.hanke@gmail.com',
          license      = 'MIT License',
          url          = 'https://github.com/neurodebian/testkraut',
          download_url = 'https://github.com/neurodebian/testkraut/tags',
          description  = 'test and evaluate heterogeneous data processing pipelines',
          long_description = open('README.rst').read(),
          classifiers  = ["Development Status :: 3 - Alpha",
                          "Environment :: Console",
                          "Intended Audience :: Science/Research",
                          "License :: OSI Approved :: MIT License",
                          "Operating System :: OS Independent",
                          "Programming Language :: Python",
                          "Topic :: Scientific/Engineering"],
          platforms    = "OS Independent",
          provides     = ['testkraut'],
          # please maintain alphanumeric order
          packages     = [ 'testkraut',
                           'testkraut.cmdline',
                           'testkraut.evaluators',
                           'testkraut.external',
                           'testkraut.tests',
                           ],
          package_data = {
              'testkraut': test_library_files + ['testkraut.cfg']
            },
          scripts      = glob(os.path.join('bin', '*'))
          )

if __name__ == "__main__":
    main(**extra_setuptools_args)
